from marqo.neural_search import enums
from typing import Iterable, Container
from marqo.errors import MarqoError
from marqo.neural_search.enums import NeuralField
from marqo.neural_search import constants

def validate_field_name(field_name) -> str:
    """TODO:
        - length (remember the vector name will have the vector_prefix added to the front of field_name)
        - consider blanket "no double names starting with double underscore..."
    Args:
        field_name:

    returns field_name, if all validations pass

    Raises:
        MarqoError
    """
    if not isinstance(field_name, str):
        raise MarqoError(F"field name must be str! Found type {type(field_name)} for {field_name}")
    if not field_name:
        raise MarqoError("field name can't be empty! ")
    if field_name.startswith(enums.NeuralField.vector_prefix):
        raise MarqoError(F"can't start field name with protected prefix {enums.NeuralField.vector_prefix}."
                            F" Error raised for field name: {field_name}")
    if field_name.startswith(enums.NeuralField.chunks):
        raise MarqoError(F"can't name field with protected field name {enums.NeuralField.chunks}."
                            F" Error raised for field name: {field_name}")
    char_validation = [(c, c not in constants.ILLEGAL_CUSTOMER_FIELD_NAME_CHARS)
                        for c in field_name]
    char_validation_failures = [c for c in char_validation if not c[1]]
    if char_validation_failures:
        raise MarqoError(F"Illegal character '{char_validation_failures[0][0]}' "
                            F"detected in field name {field_name}")
    if field_name not in enums.NeuralField.__dict__.values():
        return field_name
    else:
        raise MarqoError(f"field name can't be a protected field. Please rename this field: {field_name}")


def validate_doc(doc: dict) -> dict:
    """
    Args:
        doc: a document indexed by the client

    Raises an MarqoError

    Returns
        doc if all validations pass
    """
    if len(doc) <= 0:
        raise MarqoError("Can't index empty dict!")
    return doc


def validate_vector_name(name: str):
    """Checks that the vector name is valid.
    It should have the form __vector_{customer field name}
    """
    if not isinstance(name, str):
        raise MarqoError(F"vector name must be str! Found type {type(name)} for {name}")
    if not name:
        raise MarqoError("vector name can't be empty! ")

    if not name.startswith(enums.NeuralField.vector_prefix):
        raise MarqoError(
            f"Names of vectors must begin "
            f"with the vector prefix ({enums.NeuralField.vector_prefix})! \n"
            f"The name of the vector that raised the error: {name}")
    without_prefix = name.replace(enums.NeuralField.vector_prefix, '', 1)
    if not without_prefix:
        raise MarqoError(
            f"Vector name without prefix cannot be empty. "
            f"The name of the vector that raised the error: {name}"
        )
    if without_prefix in enums.NeuralField.__dict__.values():
        raise MarqoError(
            f"Vector name without vector prefix can't be a protected name."
            f"The name of the vector that raised the error: {name}"
        )
    if without_prefix == '_id':
        raise MarqoError(
            f"Vector name without vector prefix can't be a protected name."
            f"The name of the vector that raised the error: {name}"
        )
    return name


def validate_searchable_vector_props(existing_vector_properties: Container[str],
                               subset_vector_properties: Iterable[str]) -> Iterable[str]:
    """Validates that the a subset of vector properties is indeed a subset.

    Args:
        existing_vector_properties: assumes that each name begins with the vector prefix
        subset_vector_properties: assumes that each name begins with the vector prefix

    Returns:
        subset_vector_properties if validation has passed
    Raises
        S2Search error in case where subset_vector_properties isn't a subset of
        existing_vector_properties

    """
    for subset_vec in subset_vector_properties:
        if subset_vec not in existing_vector_properties:
            raise MarqoError(f"Searchable attribute '{subset_vec.replace(NeuralField.vector_prefix, '')}' "
                                f"not found in index.")
    return subset_vector_properties



# from torch import FloatTensor
# from typing import Any, Dict, List, Optional, Union
from pathlib import Path
import os

import numpy as np
from PIL import Image
import torch
import torch.nn.functional as F
from torch import nn
import onnxruntime
import onnx

from marqo.s2_inference.types import *
from marqo.s2_inference.clip_utils import (
    format_and_load_CLIP_image,
    format_and_load_CLIP_images,
    _is_image,
)
from marqo.s2_inference.sbert_onnx_utils import ModelCache
from marqo.s2_inference.logger import get_logger
logger = get_logger(__name__)

def normalize_2d(inputs):

    is_valid = False
    if isinstance(inputs, FloatTensor):
        n_dims = inputs.dim()
        if n_dims == 2:
            row_sums = inputs.norm(dim=-1, keepdim=True)
            is_valid = True
    elif isinstance(inputs, ndarray):
        n_dims = inputs.ndim
        if n_dims == 2:
            row_sums = np.linalg.norm(inputs, axis=1, ord=2)[:, np.newaxis]
            is_valid = True
    elif isinstance(inputs, list):
        return normalize_2d(np.array(inputs))
    else:
        raise TypeError(f"unrecognized type {type(inputs)}")

    if is_valid:
        return inputs / row_sums

    raise TypeError(f"expected 2D matrix for normalization but received {n_dims}")  
    


class Textual(nn.Module):
    """https://colab.research.google.com/drive/1YqwLsBEP2qn3M_tgqXIKCYEjtVfFoKMy#scrollTo=WkAB99F-pneX

    Args:
        nn (_type_): _description_
    """
    def __init__(self, model):
        super().__init__()
        self.transformer = model.transformer
        self.positional_embedding = model.positional_embedding
        self.transformer = model.transformer
        self.ln_final = model.ln_final
        self.text_projection = model.text_projection
        self.token_embedding = model.token_embedding

    def forward(self, text):
        x = self.token_embedding(text)  # [batch_size, n_ctx, d_model]

        x = x + self.positional_embedding
        x = x.permute(1, 0, 2)  # NLD -> LND
        x = self.transformer(x)
        x = x.permute(1, 0, 2)  # LND -> NLD
        x = self.ln_final(x)

        # x.shape = [batch_size, n_ctx, transformer.width]
        # take features from the eot embedding (eot_token is the highest number in each sequence)
        # needs .float() before .argmax(  ) to work
        x = x[torch.arange(x.shape[0]), text.float().argmax(dim=-1)] @ self.text_projection

        return x


def attention(self, x: torch.Tensor):
    # onnx doesn't like multi_head_attention_forward so this is a reimplementation
    q, k, v = (torch.einsum("tbh, oh -> tbo", x, self.attn.in_proj_weight) + self.attn.in_proj_bias).contiguous().chunk(3, dim=-1)
    tgt_len = q.shape[0]
    bsz = q.shape[1]
    num_heads = self.attn.num_heads
    head_dim = q.shape[2] // num_heads
    attn_output, attn_output_weights = F._scaled_dot_product_attention(
        q.reshape(tgt_len, bsz * num_heads, head_dim).transpose(0, 1),
        k.reshape(tgt_len, bsz * num_heads, head_dim).transpose(0, 1),
        v.reshape(tgt_len, bsz * num_heads, head_dim).transpose(0, 1), None, 0.0
        )
    attn_output = attn_output.transpose(0, 1).contiguous().view(q.shape)
    attn_output = F.linear(attn_output, self.attn.out_proj.weight, self.attn.out_proj.bias)
    return attn_output

class ConvertCLIP:

    default_onnx_params = dict(input_names=['input'], output_names=['output'],
                      export_params=True, verbose=False, opset_version=12,
                      do_constant_folding=True,
                      dynamic_axes={'input': {0: 'batch_size'}, 'output': {0: 'batch_size'}})

    def __init__(self, model_name_or_path: Optional[str] = None,
                 cache_folder: Optional[str] = None,
                 onnx_folder: Optional[str] = None,
                 vision_onnx_model_name: Optional[str] = None,
                 text_onnx_model_name: Optional[str] = None,
                 enable_overwrite: Optional[bool] = False,
                 ):

        # some paths
        self.cache_folder = cache_folder
        self.onnx_folder = onnx_folder
        self.vision_onnx_model_name = vision_onnx_model_name
        self.text_onnx_model_name = text_onnx_model_name
        self.model_name_or_path = model_name_or_path

        # should we overwrite even if another exists
        self.enable_overwrite = enable_overwrite

        # final names for the export onnx models
        self.vision_export_model_name = None
        self.text_export_model_name = None

        # have a flag to see if either exists or not
        self.vision_already_exists = None
        self.text_already_exists = None

        # dummy inputs
        self.text_dummy_input = None
        self.text_dummy_input_tokenized = None
        self.vision_dummy_input = None

        #
        self.model = None
        self.preprocess = None
        self.tokenizer = None

        logger.info("getting model paths...")
        self._get_paths()
        logger.info("checking if onnx already exists...")
        self._check_exists()

        logger.info('loading original clip...')
        self._load_original_clip()
        
        if not self.vision_already_exists or not self.text_already_exists:
            logger.info('done...')
            logger.info("getting dummy inputs for conversion...")
            self._get_dummy_inputs()
            logger.info('done...')


    def _load_original_clip(self):
        import clip
        #clip.model.ResidualAttentionBlock.attention = attention
        self.model, self.preprocess = clip.load(self.model_name_or_path, device="cpu", jit=False)
        self.tokenizer = clip.tokenize
        self.model.eval()

    def _check_exists(self):

        # construct all the save paths and cache directories
        if self.vision_export_model_name is None or self.text_export_model_name is None:
            self._get_paths()

        # if doesn't exist or overwrite, then convert
        if not os.path.isfile(self.vision_export_model_name) or self.enable_overwrite:
            self.vision_already_exists = False
        else:
            self.vision_already_exists = True

        # if doesn't exist or overwrite, then convert
        if not os.path.isfile(self.text_export_model_name) or self.enable_overwrite:
            self.text_already_exists = False
        else:
            self.text_already_exists = True


    def _get_paths(self) -> None:
        """get the paths of the cache, onnx save path and output model path
        """
        if self.onnx_folder is None:
            self.onnx_folder = ModelCache.onnx_cache_path
            Path(self.onnx_folder).mkdir(parents=True, exist_ok=True)

        if self.cache_folder is None:
            self.cache_folder = ModelCache.torch_cache_path

        if self.vision_onnx_model_name is None:
            self.vision_onnx_model_name = f"{os.path.basename(self.model_name_or_path.replace('/', '_'))}_vision.onnx"

        if self.text_onnx_model_name is None:
            self.text_onnx_model_name = f"{os.path.basename(self.model_name_or_path.replace('/', '_'))}_text.onnx"

        self.vision_export_model_name = os.path.join(self.onnx_folder, f"{self.vision_onnx_model_name}") 
        self.text_export_model_name = os.path.join(self.onnx_folder, f"{self.text_onnx_model_name}") 

    def _get_dummy_inputs(self):

        if self.text_dummy_input is None:
            self.text_dummy_input = ['bingo', 'bango', 'dingo']
            self.text_dummy_input_tokenized = self.tokenizer(self.text_dummy_input).cpu()

        if self.vision_dummy_input is None:
            self.vision_dummy_input = [Image.new("RGB", (600,600), (255,255,255)), Image.new("RGB", (600,600), (255,255,255))]
            self.vision_dummy_input_processed = torch.stack([self.preprocess(_i).cpu() for _i in self.vision_dummy_input])

    def convert(self):

        if not self.text_already_exists:
            self.convert_text()
            self.text_already_exists = True
        if not self.vision_already_exists:
            self.convert_vision()
            self.vision_already_exists = True

    def convert_vision(self):

        if self.vision_dummy_input is None:
            self._get_dummy_inputs()

        visual = self.model.visual
        self.torch_export(visual, self.vision_dummy_input_processed,
                            self.vision_export_model_name, self.default_onnx_params)

        self.onnx_checker(self.vision_export_model_name)

    def convert_text(self):
        textual = Textual(self.model)

        if self.text_dummy_input is None:
            self._get_dummy_inputs()
        self.torch_export(textual, self.text_dummy_input_tokenized, self.text_export_model_name,
                          export_params=self.default_onnx_params)
        self.onnx_checker(self.text_export_model_name)

    @staticmethod
    def torch_export(model, dummy_input, path: str, export_params: dict):
        torch.onnx.export(model, dummy_input, path, **export_params)

    @staticmethod
    def onnx_checker(path: str):
        model = onnx.load(path)
        onnx.checker.check_model(model)
        del model

    def clean_up(self):
        if self.model is not None:
            del self.model
        
class CLIP_ONNX(ConvertCLIP):
    
    """
    conveniance class wrapper to make clip work easily for both text and image encoding
    """

    def __init__(self, model_type: str = "ViT-B/32", device: str = 'cpu',  
                            embedding_dim: int = None, 
                            cache_folder: Optional[str] = None,
                            onnx_folder: Optional[str] = None,
                            vision_onnx_model_name: Optional[str] = None,
                            text_onnx_model_name: Optional[str] = None,
                            enable_overwrite: Optional[bool] = False,
                            **kwargs) -> None:
        super().__init__(model_type,
                 cache_folder,
                 onnx_folder,
                 vision_onnx_model_name,
                 text_onnx_model_name,
                 enable_overwrite)

        self.convert()
        self.clean_up()

        self.model_type = model_type
        self.device = device
        self.model = None
        self.embedding_dimension = embedding_dim

    def load(self):
        self._get_onnx_provider()
        self._load_clip_session()

    def _get_onnx_provider(self) -> None:
        """determine where the model should run based on specified device
        """
        self.onnxproviders = onnxruntime.get_available_providers()
        logger.info(f"device:{self.device}")
        if self.device == 'cpu':
            self.fast_onnxprovider = 'CPUExecutionProvider'
        else:
            if 'CUDAExecutionProvider' not in self.onnxproviders:
                self.fast_onnxprovider = 'CPUExecutionProvider'
            else:
                self.fast_onnxprovider = 'CUDAExecutionProvider'

        logger.info(f"onnx_provider:{self.fast_onnxprovider}")

    def _load_clip_session(self):
        self.vision_session = onnxruntime.InferenceSession(self.vision_export_model_name,
                                                               providers=[self.fast_onnxprovider])
        self.text_session = onnxruntime.InferenceSession(self.text_export_model_name,
                                                                providers=[self.fast_onnxprovider])

    def _encode(self, input, modality):

        if modality == 'text':
            session = self.text_session        
        elif modality == 'vision':
            session = self.vision_session
        else:
            raise ValueError(f"expected modailty to be 'text' or 'vision' but received {modality}")

        onnx_input = {session.get_inputs()[0].name: input}
        output, = session.run(None, onnx_input)
        return output

    def encode_image(self, images, normalize = True):

        # default to batch encoding
        if isinstance(images, list):
            image_input = format_and_load_CLIP_images(images)
        else:
            image_input = [format_and_load_CLIP_image(images)]

        # image_input_processed = {self.vision_session.get_inputs()[0].name:self.preprocess(_img).to(self.device)} for _img in image_input]
        self.image_input_processed = np.array([self.preprocess(_img).cpu().numpy() for _img in image_input])

        if normalize:
            return normalize_2d(self._encode(self.image_input_processed, 'vision'))
        else:
            return self._encode(self.image_input_processed, 'vision')

    def encode_text(self, text, normalize = True):
        tokenized_text = self.tokenizer(text).detach().cpu().numpy().astype(np.int32)
        if normalize:
            return normalize_2d(self._encode(tokenized_text, 'text'))
        else:
            return self._encode(tokenized_text, 'text')

    def encode(self, inputs: Union[str, ImageType, List[Union[str, ImageType]]], 
                default: str = 'text', normalize = True, **kwargs) -> FloatTensor:

        infer = kwargs.pop('infer', True)

        if infer and _is_image(inputs):
            is_image = True
        else:
            is_image = False
            if default == 'text':
                is_image = False
            elif default == 'image':
                is_image = True
            else:
                raise ValueError(f"expected default='image' or default='text' but received {default}")

        if is_image:
            print("encoding image")
            return self.encode_image(inputs, normalize=normalize)
        else:
            print("encoding text")
            return self.encode_text(inputs, normalize=normalize)



import base64
import hashlib
import hmac
import json
import datetime
import nltk
import ssl
from typing import Any, Dict, List, Optional, Union
from marqo.index import Index
from marqo.enums import Devices
from marqo.config import Config
from marqo._httprequests import HttpRequests
from marqo.errors import MarqoError
from marqo.neural_search import neural_search, index_meta_cache
import urllib3
from marqo import utils, enums


class Client:
    """
    A client for the marqo API

    A client instance is needed for every marqo API method to know the location of
    marqo and its permissions.
    """



    def __init__(
        self, url: str = "http://localhost:9200", main_user: str = None, main_password: str = None,
        indexing_device: Optional[Union[enums.Devices, str]] = None,
        search_device: Optional[Union[enums.Devices, str]] = None
    ) -> None:
        """
        Parameters
        ----------
        url:
            The url to the S2Search API (ex: http://localhost:9200)
        """
        self._ensure_nltk_setup()

        self.main_user = main_user
        self.main_password = main_password
        if (main_user is not None) and (main_password is not None):
            self.url = utils.construct_authorized_url(url_base=url, username=main_user, password=main_password)
        else:
            self.url = url
        self.config = Config(self.url, indexing_device=indexing_device, search_device=search_device)
        self.http = HttpRequests(self.config)
        index_meta_cache.populate_cache(config=self.config)
        self._turn_off_auto_create_index()

    def _turn_off_auto_create_index(self):
        """turns off auto creation of indices. To be run in client instantiation"""
        self.http.put(
            path="_cluster/settings",
            body={
                "persistent": {"action.auto_create_index": "false"}
            })


    def create_index(
        self, index_name: str,
        treat_urls_and_pointers_as_images=False, model=None,
        normalize_embeddings=True,
        sentences_per_chunk=2,
        sentence_overlap=0
    ) -> Dict[str, Any]:
        """Create an index.

        Parameters
        ----------
        index_name: str
            name of the index.


        Returns
        -------
        task:
            Name of the index

        Raises
        ------
        s2SearchApiError
            An error containing details about why marqo can't process your request. marqo error codes are described here: https://docs.marqo.com/errors/#marqo-errors
        """
        return Index.create(
            config=self.config, index_name=index_name,
            treat_urls_and_pointers_as_images=treat_urls_and_pointers_as_images,
            model=model, normalize_embeddings=normalize_embeddings,
            sentences_per_chunk=sentences_per_chunk, sentence_overlap=sentence_overlap
        )

    def delete_index(self, index_name: str) -> Dict[str, Any]:
        """Deletes an index

        Parameters
        ----------
        index_name:
            UID of the index.

        Returns
        -------
        task:
            Dictionary containing a task to track the informations about the progress of an asynchronous process.
            https://docs.marqo.com/reference/api/tasks.html#get-one-task

        Raises
        ------
        s2SearchApiError
            An error containing details about why marqo can't process your request. marqo error codes are described here: https://docs.marqo.com/errors/#marqo-errors
        """

        return neural_search.delete_index(config=self.config, index_name=index_name)

    def get_index(self, index_name: str) -> Index:
        """Get the index.
        This index should already exist.

        Parameters
        ----------
        index_name:
            UID of the index.

        Returns
        -------
        index:
            An Index instance containing the information of the fetched index.

        Raises
        ------
        s2SearchApiError
            An error containing details about why marqo can't process your request. marqo error codes are described here: https://docs.marqo.com/errors/#marqo-errors
        """
        ix = Index(self.config, index_name)
        # verify it exists
        self.http.get(path=index_name)
        return ix

    def index(self, index_name: str) -> Index:
        """Create a local reference to an index identified by UID, without doing an HTTP call.
        Calling this method doesn't create an index in the marqo instance, but grants access to all the other methods in the Index class.

        Parameters
        ----------
        index_name:
            UID of the index.

        Returns
        -------
        index:
            An Index instance.
        """
        if index_name is not None:
            return Index(self.config, index_name=index_name)
        raise Exception('The index UID should not be None')

    def get_version(self) -> Dict[str, str]:
        """Get version marqo

        Returns
        -------
        version:
            Information about the version of marqo.

        Raises
        ------
        s2SearchApiError
            An error containing details about why marqo can't process your request. marqo error codes are described here: https://docs.marqo.com/errors/#marqo-errors
        """
        return self.http.get(self.config.paths.version)

    def version(self) -> Dict[str, str]:
        """Alias for get_version

        Returns
        -------
        version:
            Information about the version of marqo.

        Raises
        ------
        s2SearchApiError
            An error containing details about why marqo can't process your request. marqo error codes are described here: https://docs.marqo.com/errors/#marqo-errors
        """
        return self.get_version()

    @staticmethod
    def _base64url_encode(
        data: bytes
    ) -> str:
        return base64.urlsafe_b64encode(data).decode('utf-8').replace('=','')

    def _ensure_nltk_setup(self):
        try:
            nltk.data.find('tokenizers/punkt')
        except LookupError:
            nltk.download('punkt')



"""This is the interface for interacting with S2 Inference
The functions defined here would have endpoints, later on.
"""
import numpy as np

from marqo.s2_inference.model_registry import load_model_properties
from marqo.s2_inference.configs import get_default_device,get_default_normalization,get_default_seq_length
from marqo.s2_inference.types import *
from marqo.s2_inference.logger import get_logger
logger = get_logger(__name__)

available_models = dict()
MODEL_PROPERTIES = load_model_properties()

def vectorise(model_name: str, content: Union[str, List[str]], device: str = get_default_device(), 
                                    normalize_embeddings: bool = get_default_normalization(), **kwargs) -> List[List[float]]:
    """vectorizes the content by model name

    Args:
        model_name (str): _description_
        content (_type_): _description_

    Returns:
        List[List[float]]: _description_
    """
    # TODO map by (model_name, device) instead of model_name
    if model_name not in available_models:
        available_models[model_name] = _load_model(model_name, device=device)
        logger.info(f'loaded {model_name} on device {device} with normalization={normalize_embeddings}')
    vectorised = available_models[model_name].encode(content, normalize=normalize_embeddings, **kwargs)
    return _convert_vectorized_output(vectorised)


def clear_loaded_models() -> None:
    """ clears the loaded model cache
    """
    available_models.clear()


def get_model_properties(model_name: str) -> dict:
    """ Returns a dict describing properties of a model.

    These properties will be used by the Neural_Search application to set up
    index parameters.

    see https://huggingface.co/sentence-transformers for available models

    TODO: standardise these dicts

    Returns:
        dict: a dictionary describing properties of the model.
    """

    if model_name not in MODEL_PROPERTIES['models']:
        raise KeyError(f"model_name={model_name} not in allowed model names")

    return MODEL_PROPERTIES['models'][model_name]


def _check_output_type(output: List[List[float]]) -> bool:
    """checks the output type conforms to what we want

    Args:
        output (List[List[float]]): _description_

    Returns:
        bool: _description_
    """

    if not isinstance(output, list):
        return False
    elif len(output) == 0:
        raise ValueError("received empty input")

    if not isinstance(output[0], list):
        return False
    elif len(output[0]) == 0:
        raise ValueError("received empty input")
    
    # this is a soft check for speed reasons
    if not isinstance(output[0][0], (float, int)):
        return False

    return True


def _float_tensor_to_list(output: FloatTensor, device: str = get_default_device()) -> Union[List[List[float]], List[float]]:
    """

    Args:
        output (FloatTensor): _description_

    Returns:
        List[List[float]]: _description_
    """

    return output.detach().to(device).tolist()


def _nd_array_to_list(output : ndarray) -> Union[List[List[float]], List[float]]:
    """

    Args:
        output (ndarray): _description_

    Returns:
        List[List[float]]: _description_
    """

    return output.tolist()


def _convert_vectorized_output(output: Union[FloatTensor, ndarray, List[List[float]]], fp16: bool = False) -> List[List[float]]:
    """converts the model outputs to a list of lists of floats
    also checks the input dim, expects (samples x vector dim)
    if a single sample is present, will pad the first dim to make it (1 x vector_dim)

    Args:
        output (FloatTensor): _description_

    Returns:
        List[List[float]]: _description_
    """

    if _check_output_type(output):
        return output

    if isinstance(output, (FloatTensor, Tensor)):
        if output.ndim == 1:
            output = output.unsqueeze(0)        
        output = _float_tensor_to_list(output)
    
    elif isinstance(output, ndarray):
        if output.ndim == 1:
            output = output[np.newaxis, :]

        output = _nd_array_to_list(output)

    elif isinstance(output, list):
        if isinstance(output[0], FloatTensor):
            output = [_float_tensor_to_list(_o) for _o in output]
        elif isinstance(output[0], ndarray):
            output = [_nd_array_to_list(_o) for _o in output]
        else:
            raise TypeError(f"unsupported nested list with elements of type {type(output[0])}")

    else:
        raise TypeError(f"unsupported output type of {type(output)}")

    if fp16:
        output = np.array(output).astype(np.float16).tolist()

    if _check_output_type(output):
        return output
    
    raise TypeError(f"unable to convert input of type {type(output)} to a list of lists of floats")


def _get_model_loader(model_name: str) -> Any:
    """ Returns a dict describing properties of a model.

    These properties will be used by the Neural_Search application to set up
    index parameters.

    see https://huggingface.co/sentence-transformers for available models

    TODO: standardise these dicts

    Returns:
        dict: a dictionary describing properties of the model.
    """

    if model_name not in MODEL_PROPERTIES['models']:
        raise KeyError(f"model_name={model_name} not in allowed model names")

    model_type = MODEL_PROPERTIES['models'][model_name]['type']

    if model_type not in MODEL_PROPERTIES['loaders']:
        raise KeyError(f"model_name={model_name} for model_type={model_type} not in allowed model types")

    return MODEL_PROPERTIES['loaders'][model_type]


def _load_model(model_name: str, device: str = get_default_device()) -> Any:
    """_summary_

    Args:
        model_name (str): _description_
        device (str, optional): _description_. Defaults to 'cpu'.

    Returns:
        Any: _description_
    """

    model_properties = get_model_properties(model_name)
    
    loader = _get_model_loader(model_name)
    
    max_sequence_length = model_properties.get('tokens', get_default_seq_length())

    model = loader(model_properties['name'], device=device, embedding_dim=model_properties['dimensions'], 
                    max_seq_length=max_sequence_length)
    
    model.load()

    return model

# def normalize(inputs):

#     is_valid = False
#     if isinstance(inputs, FloatTensor):
#         n_dims = inputs.dim()
#         if n_dims == 2:
#             row_sums = inputs.norm(dim=-1, keepdim=True)
#             is_valid = True
#     elif isinstance(inputs, ndarray):
#         n_dims = inputs.ndim
#         if n_dims == 2:
#             row_sums = np.linalg.norm(inputs, axis=1, ord=2)[:, np.newaxis]
#             is_valid = True
#     elif isinstance(inputs, list):
#         return normalize(np.array(inputs))
#     else:
#         raise TypeError(f"unrecognized type {type(inputs)}")

#     if is_valid:
#         return inputs / row_sums

#     raise TypeError(f"expected 2D matrix for normalization but received {n_dims}")  
    



from marqo.s2_inference.hf_utils import HF_MODEL
from marqo.s2_inference.sbert_onnx_utils import SBERT_ONNX
from marqo.s2_inference.sbert_utils import SBERT, TEST
from marqo.s2_inference.random_utils import Random
from marqo.s2_inference.clip_utils import CLIP
from marqo.s2_inference.clip_onnx_utils import CLIP_ONNX
from marqo.s2_inference.types import Any, Dict, List, Optional, Union, FloatTensor

# we need to keep track of the embed dim and model load functions/classes
# we can use this as a registry 
def _get_clip_properties() -> Dict:
    CLIP_MODEL_PROPERTIES = {
            'RN50':
                {"name": "RN50",
                "dimensions": 1024,
                "notes": "CLIP resnet50",
                "type": "clip",
                },
            'RN101':
                {"name": "RN101",
                "dimensions": 512,
                "notes": "CLIP resnet101",
                "type": "clip",
                },
            'RN50x4':
                {"name": "RN50x4",
                "dimensions": 640,
                "notes": "CLIP resnet50x4",
                "type": "clip",
                },              
            'RN50x16':
                {"name": "RN50x16",
                "dimensions": 768,
                "notes": "CLIP resnet50x16",
                "type": "clip",
                },
            'RN50x64':
                {"name": "RN50x64",
                "dimensions": 1024,
                "notes": "CLIP resnet50x64",
                "type": "clip",
                },
            'ViT-B/32':
                {"name": "ViT-B/32",
                "dimensions": 512,
                "notes": "CLIP ViT-B/32",
                "type": "clip",
                },
            'ViT-B/16':
                {"name": "ViT-B/16",
                "dimensions": 512,
                "notes": "CLIP ViT-B/16",
                "type":"clip",
                },
            'ViT-L/14':
                {"name": "ViT-L/14",
                "dimensions": 768,
                "notes": "CLIP ViT-L/14",
                "type":"clip",
                },
            'ViT-L/14@336px':
                {"name": "ViT-L/14@336px",
                "dimensions": 768,
                "notes": "CLIP ViT-L/14@336px",
                "type":"clip",
                },

        }
    return CLIP_MODEL_PROPERTIES

def _get_clip_onnx_properties() -> Dict:
    CLIP_ONNX_MODEL_PROPERTIES = {
            'onnx/RN50':
                {"name": "RN50",
                "dimensions": 1024,
                "notes": "CLIP resnet50",
                "type": "clip_onnx",
                },
            'onnx/RN101':
                {"name": "RN101",
                "dimensions": 512,
                "notes": "CLIP resnet101",
                "type": "clip_onnx",
                },
            'onnx/RN50x4':
                {"name": "RN50x4",
                "dimensions": 640,
                "notes": "CLIP resnet50x4",
                "type": "clip_onnx",
                },              
            'onnx/RN50x16':
                {"name": "RN50x16",
                "dimensions": 768,
                "notes": "CLIP resnet50x16",
                "type": "clip_onnx",
                },
            'onnx/RN50x64':
                {"name": "RN50x64",
                "dimensions": 1024,
                "notes": "CLIP resnet50x64",
                "type": "clip_onnx",
                },
            'onnx/ViT-B/32':
                {"name": "ViT-B/32",
                "dimensions": 512,
                "notes": "CLIP ViT-B/32",
                "type": "clip_onnx",
                },
            'onnx/ViT-B/16':
                {"name": "ViT-B/16",
                "dimensions": 512,
                "notes": "CLIP ViT-B/16",
                "type":"clip_onnx",
                },
            'onnx/ViT-L/14':
                {"name": "ViT-L/14",
                "dimensions": 768,
                "notes": "CLIP ViT-L/14",
                "type":"clip_onnx",
                },
            'onnx/ViT-L/14@336px':
                {"name": "ViT-L/14@336px",
                "dimensions": 768,
                "notes": "CLIP ViT-L/14@336px",
                "type":"clip_onnx",
                },

        }
    return CLIP_ONNX_MODEL_PROPERTIES

def _get_sbert_properties() -> Dict:
    SBERT_MODEL_PROPERTIES = {
            "sentence-transformers/all-MiniLM-L6-v1": 
                {"name": "sentence-transformers/all-MiniLM-L6-v1",
                "dimensions": 384,
                "tokens":128,
                "type":"sbert",
                "notes": ""},
            "sentence-transformers/all-MiniLM-L6-v2": 
                {"name": "sentence-transformers/all-MiniLM-L6-v2",
                "dimensions": 384,
                "tokens":256,
                "type":"sbert",
                "notes": ""},
            "sentence-transformers/all-mpnet-base-v1": 
                {"name": "sentence-transformers/all-mpnet-base-v1",
                "dimensions": 768,
                "tokens":128,
                "type":"sbert",
                "notes": ""},
            "sentence-transformers/all-mpnet-base-v2": 
                {"name": "sentence-transformers/all-mpnet-base-v2",
                "dimensions": 768,
                "tokens":128,
                "type":"sbert",
                "notes": ""},

            "flax-sentence-embeddings/all_datasets_v3_MiniLM-L12": 
                {"name": "flax-sentence-embeddings/all_datasets_v3_MiniLM-L12",
                "dimensions": 384,
                "tokens":128,
                "type":"sbert",
                "notes": ""},
            "flax-sentence-embeddings/all_datasets_v3_MiniLM-L6": 
                {"name": "flax-sentence-embeddings/all_datasets_v3_MiniLM-L6",
                "dimensions": 384,
                "tokens":128,
                "type":"sbert",
                "notes": ""},
            "flax-sentence-embeddings/all_datasets_v4_MiniLM-L12": 
                {"name": "flax-sentence-embeddings/all_datasets_v4_MiniLM-L12",
                "dimensions": 384,
                "tokens":128,
                "type":"sbert",
                "notes": ""},
            "flax-sentence-embeddings/all_datasets_v4_MiniLM-L6": 
                {"name": "flax-sentence-embeddings/all_datasets_v4_MiniLM-L6",
                "dimensions": 384,
                "tokens":128,
                "type":"sbert",
                "notes": ""},

            "flax-sentence-embeddings/all_datasets_v3_mpnet-base": 
                {"name": "flax-sentence-embeddings/all_datasets_v3_mpnet-base",
                "dimensions": 768,
                "tokens":128,
                "type":"sbert",
                "notes": ""},
            "flax-sentence-embeddings/all_datasets_v4_mpnet-base": 
                {"name": "flax-sentence-embeddings/all_datasets_v4_mpnet-base",
                "dimensions": 768,
                "tokens":128,
                "type":"sbert",
                "notes": ""},
    }
    return SBERT_MODEL_PROPERTIES

def _get_hf_properties() -> Dict:
    HF_MODEL_PROPERTIES = {
            "hf/all-MiniLM-L6-v1": 
                {"name": "sentence-transformers/all-MiniLM-L6-v1",
                "dimensions": 384,
                "tokens":128,
                "type":"hf",
                "notes": ""},
            "hf/all-MiniLM-L6-v2": 
                {"name": "sentence-transformers/all-MiniLM-L6-v2",
                "dimensions": 384,
                "tokens":256,
                "type":"hf",
                "notes": ""},
            "hf/all-mpnet-base-v1": 
                {"name": "sentence-transformers/all-mpnet-base-v1",
                "dimensions": 768,
                "tokens":128,
                "type":"hf",
                "notes": ""},
            "hf/all-mpnet-base-v2": 
                {"name": "sentence-transformers/all-mpnet-base-v2",
                "dimensions": 768,
                "tokens":128,
                "type":"hf",
                "notes": ""},

            "hf/all_datasets_v3_MiniLM-L12": 
                {"name": "flax-sentence-embeddings/all_datasets_v3_MiniLM-L12",
                "dimensions": 384,
                "tokens":128,
                "type":"hf",
                "notes": ""},
            "hf/all_datasets_v3_MiniLM-L6": 
                {"name": "flax-sentence-embeddings/all_datasets_v3_MiniLM-L6",
                "dimensions": 384,
                "tokens":128,
                "type":"hf",
                "notes": ""},
            "hf/all_datasets_v4_MiniLM-L12": 
                {"name": "flax-sentence-embeddings/all_datasets_v4_MiniLM-L12",
                "dimensions": 384,
                "tokens":128,
                "type":"hf",
                "notes": ""},
            "hf/all_datasets_v4_MiniLM-L6": 
                {"name": "flax-sentence-embeddings/all_datasets_v4_MiniLM-L6",
                "dimensions": 384,
                "tokens":128,
                "type":"hf",
                "notes": ""},

            "hf/all_datasets_v3_mpnet-base": 
                {"name": "flax-sentence-embeddings/all_datasets_v3_mpnet-base",
                "dimensions": 768,
                "tokens":128,
                "type":"hf",
                "notes": ""},
            "hf/all_datasets_v4_mpnet-base": 
                {"name": "flax-sentence-embeddings/all_datasets_v4_mpnet-base",
                "dimensions": 768,
                "tokens":128,
                "type":"hf",
                "notes": ""},
    }
    return HF_MODEL_PROPERTIES

def _get_sbert_onnx_properties() -> Dict:
    SBERT_ONNX_MODEL_PROPERTIES = {
            "onnx/all-MiniLM-L6-v1": 
                {"name": "sentence-transformers/all-MiniLM-L6-v1",
                "dimensions": 384,
                "tokens":128,
                "type":"sbert_onnx",
                "notes": ""},
            "onnx/all-MiniLM-L6-v2": 
                {"name": "sentence-transformers/all-MiniLM-L6-v2",
                "dimensions": 384,
                "tokens":256,
                "type":"sbert_onnx",
                "notes": ""},
            "onnx/all-mpnet-base-v1": 
                {"name": "sentence-transformers/all-mpnet-base-v1",
                "dimensions": 768,
                "tokens":128,
                "type":"sbert_onnx",
                "notes": ""},
            "onnx/all-mpnet-base-v2": 
                {"name": "sentence-transformers/all-mpnet-base-v2",
                "dimensions": 768,
                "tokens":128,
                "type":"sbert_onnx",
                "notes": ""},

            "onnx/all_datasets_v3_MiniLM-L12": 
                {"name": "flax-sentence-embeddings/all_datasets_v3_MiniLM-L12",
                "dimensions": 384,
                "tokens":128,
                "type":"sbert_onnx",
                "notes": ""},
            "onnx/all_datasets_v3_MiniLM-L6": 
                {"name": "flax-sentence-embeddings/all_datasets_v3_MiniLM-L6",
                "dimensions": 384,
                "tokens":128,
                "type":"sbert_onnx",
                "notes": ""},
            "onnx/all_datasets_v4_MiniLM-L12": 
                {"name": "flax-sentence-embeddings/all_datasets_v4_MiniLM-L12",
                "dimensions": 384,
                "tokens":128,
                "type":"sbert_onnx",
                "notes": ""},
            "onnx/all_datasets_v4_MiniLM-L6": 
                {"name": "flax-sentence-embeddings/all_datasets_v4_MiniLM-L6",
                "dimensions": 384,
                "tokens":128,
                "type":"sbert_onnx",
                "notes": ""},

            "onnx/all_datasets_v3_mpnet-base": 
                {"name": "flax-sentence-embeddings/all_datasets_v3_mpnet-base",
                "dimensions": 768,
                "tokens":128,
                "type":"sbert_onnx",
                "notes": ""},
            "onnx/all_datasets_v4_mpnet-base": 
                {"name": "flax-sentence-embeddings/all_datasets_v4_mpnet-base",
                "dimensions": 768,
                "tokens":128,
                "type":"sbert_onnx",
                "notes": ""},
    }
    return SBERT_ONNX_MODEL_PROPERTIES

def _get_sbert_test_properties() -> Dict:
    TEST_MODEL_PROPERTIES = {
            "sentence-transformers/test": 
                {"name": "sentence-transformers/all-MiniLM-L6-v1",
                "dimensions": 16,
                "tokens":128,
                "type":"test",
                "notes": ""},
            "test": 
                {"name": "sentence-transformers/all-MiniLM-L6-v1",
                "dimensions": 16,
                "tokens":128,
                "type":"test",
                "notes": ""},
    }
    return TEST_MODEL_PROPERTIES

def _get_random_properties() -> Dict:
    RANDOM_MODEL_PROPERTIES = {
            "random": 
                {"name": "random",
                "dimensions": 384,
                "tokens":128,
                "type":"random",
                "notes": ""},
            "random/large": 
                {"name": "random/large",
                "dimensions": 768,
                "tokens":128,
                "type":"random",
                "notes": ""},
            "random/small": 
                {"name": "random/small",
                "dimensions": 32,
                "tokens":128,
                "type":"random",
                "notes": ""},
            "random/medium": 
                {"name": "random/medium",
                "dimensions": 128,
                "tokens":128,
                "type":"random",
                "notes": ""},

    }
    return RANDOM_MODEL_PROPERTIES

def _get_model_load_mappings() -> Dict:
    return {'clip':CLIP, 'sbert':SBERT, 'test':TEST, 'sbert_onnx':SBERT_ONNX,
             'random':Random, 'hf':HF_MODEL, 'clip_onnx':CLIP_ONNX}

def load_model_properties() -> Dict:
    # also truncate the name if not already
    sbert_model_properties = _get_sbert_properties()
    sbert_model_properties.update({k.split('/')[-1]:v for k,v in sbert_model_properties.items()})

    sbert_onnx_model_properties = _get_sbert_onnx_properties()

    clip_model_properties = _get_clip_properties()
    test_model_properties = _get_sbert_test_properties()
    random_model_properties = _get_random_properties()
    hf_model_properties = _get_hf_properties()
    clip_onnx_model_properties = _get_clip_onnx_properties()

    # combine the above dicts
    model_properties = dict(clip_model_properties.items())
    model_properties.update(sbert_model_properties)
    model_properties.update(test_model_properties)
    model_properties.update(sbert_onnx_model_properties)
    model_properties.update(random_model_properties)
    model_properties.update(hf_model_properties)
    model_properties.update(clip_onnx_model_properties)

    all_properties = dict()
    all_properties['models'] = model_properties

    all_properties['loaders'] = dict()
    for key,val in _get_model_load_mappings().items():
        all_properties['loaders'][key] = val

    return all_properties




from marqo.neural_search import enums as ns_enums
from marqo.neural_search.enums import NeuralSettingsField as NsFields
import marqo.enums as mq_enums
from torch import multiprocessing as mp

def get_default_neural_index_settings():
    return {
        NsFields.index_defaults: {
            NsFields.treat_urls_and_pointers_as_images: False, # only used for models that have text and vision encoders
            NsFields.model: ns_enums.MlModel.bert,
            NsFields.normalize_embeddings: True,
            NsFields.text_preprocessing: {
                NsFields.split_length: 2,
                NsFields.split_overlap: 0,
                NsFields.split_method: ns_enums.SplitMethod.sentence
            }
        }
    }

def get_max_processes():
    return {'max_processes_cpu' :6, 'max_processes_gpu': 4}

def get_threads_per_process():
    total_cpu = max(1, mp.cpu_count() - 2)
    return max(1, total_cpu//get_max_processes()['max_processes_cpu'])



import logging

def get_logger(name):
    logger = logging.getLogger(name)
    logger.setLevel(logging.INFO)

    formatter = logging.Formatter(
        "{asctime} {threadName:>11} {levelname} {message}", style='{')

    return logger


from torch import FloatTensor, Tensor
from PIL.Image import Image as ImageType
from numpy import ndarray
from typing import (
    Any, 
    Dict, 
    List, 
    Optional, 
    Union, 
    Callable,
    Tuple, 
    Iterable, 
    Type, 
    )


"""This is a client index-IndexInfo cache. This keeps a mapping of index_name <> IndexInfo
This is a band-aid solution, as this introduces state into the Neural State application.

In the future this may be stored in redis or this logic be bundled with the
index in the search DB via a plugin.
"""
import asyncio
import datetime
import time
from multiprocessing import Process, Manager
from marqo.neural_search.models.index_info import IndexInfo
from typing import Dict
from marqo import errors
from marqo.neural_search import backend
from marqo.config import Config

index_info_cache = dict()


def empty_cache():
    global index_info_cache
    index_info_cache = dict()


def get_index_info(config: Config, index_name: str) -> IndexInfo:
    """Looks for the index name in the cache.

    If it isn't found there, it will try searching the cluster

    Args:
        config:
        index_name: name of the index

    Returns:
        IndexInfo - information about the index.

    Raises:
         MarqoError if the index isn't found on the cluster
    """
    if index_name in index_info_cache:
        return index_info_cache[index_name]
    else:
        found_index_info = backend.get_index_info(config=config, index_name=index_name)
        index_info_cache[index_name] = found_index_info
        return index_info_cache[index_name]


def get_cache() -> Dict[str, IndexInfo]:
    return index_info_cache


def refresh_index(config: Config, index_name: str) -> IndexInfo:
    """function to update an index, from the cluster.

    Args:
        config:
        index_name

    Returns:

    """
    found_index_info = backend.get_index_info(config=config, index_name=index_name)
    index_info_cache[index_name] = found_index_info
    return found_index_info


def populate_cache(config: Config):
    """Identify available index names and use them to populate the cache.

    Args:
        config:
    """
    for ix_name in backend.get_cluster_indices(config=config):
        try:
            found_index_info = backend.get_index_info(config=config, index_name=ix_name)
            index_info_cache[ix_name] = found_index_info
        except errors.MarqoNonNeuralIndexError as e:
            pass





from marqo import errors


def construct_authorized_url(url_base: str, username: str, password: str) -> str:
    """
    Args:
        url_base:
        username:
        password:

    Returns:

    """
    http_sep = "://"
    if http_sep not in url_base:
        raise errors.MarqoError(f"Could not parse url: {url_base}")
    url_split = url_base.split(http_sep)
    if len(url_split) != 2:
        raise errors.MarqoError(f"Could not parse url: {url_base}")
    http_part, domain_part = url_split
    return f"{http_part}{http_sep}{username}:{password}@{domain_part}"



import os
import time
from typing import List, Dict

import torch
import numpy as np
from torch import multiprocessing as mp

from marqo.neural_search import neural_search
from marqo.marqo_logging import logger
from marqo.neural_search.configs import get_max_processes, get_threads_per_process
from marqo.neural_search import backend
from marqo.errors import MarqoApiError

try:
    mp.set_start_method('spawn', force=True)
except:
    pass

max_processes_cpu = get_max_processes()['max_processes_cpu']
max_processes_gpu = get_max_processes()['max_processes_gpu']
max_threads_per_process = get_threads_per_process()

def get_gpu_count(device: str):
    """ returns the number of gpus. if cpu is specified defaults to 0

    Args:
        device (_type_): 'cpu' or 'cuda'

    Returns:
        _type_: _description_
    """
    if device == 'cpu':
        return 0
    
    if device.startswith('cuda'):
        return torch.cuda.device_count()

    raise ValueError(f"expected on of 'cpu', 'cuda' or 'cuda:#' but received {device}")


def get_processes(device: str):
    """returns the processes available for either cpu or cuda

    Args:
        device (str): _description_

    Raises:
        ValueError: _description_

    Returns:
        _type_: _description_
    """
    if device == 'cpu':
        return max(1, min(mp.cpu_count(), max_processes_cpu))
    
    elif device.startswith('cuda'):
        assert torch.cuda.device_count() > 0, 'cannot find cuda enabled device'
        return max_processes_gpu

    raise ValueError(f"expected on of 'cpu', 'cuda' or 'cuda:#' but received {device}")

def get_device_ids(n_processes: int, device: str):
    """gets a list of device ids based on device
        e.g. get_device_ids(2, 'cpu')
        ['cpu', 'cpu']

        # single gpu machine
        e.g. get_device_ids(2, 'cuda')
        ['cuda:0', 'cuda:0']

        # two gpu machine
        e.g. get_device_ids(2, 'cuda')
        ['cuda:0', 'cuda:1']

        # two gpu machine
        e.g. get_device_ids(4, 'cuda')
        ['cuda:0', 'cuda:1', 'cuda:0', 'cuda:1']

    Args:
        n_processes (int): _description_
        device (str): _description_

    Raises:
        ValueError: _description_

    Returns:
        _type_: _description_
    """
    if device.startswith('cpu'):
        return ['cpu']*n_processes
    elif device.startswith('cuda'):
        n_gpus = get_gpu_count('cuda')
        gpu_ids = list(range(n_gpus))*n_processes
        devices = [f'cuda:{_id}' for _id in gpu_ids]
        return devices[:n_processes]

    raise ValueError(f"expected on of 'cpu', 'cuda' or 'cuda:#' but received {device}")


class IndexChunk:

    """wrapper to pass through dopcuments to be indexed to multiprocessing
    """

    def __init__(self, config=None, index_name: str = None, docs: List[Dict] = [], 
                        auto_refresh: bool = False, batch_size: int = 50, device: str = None, process_id: int = 0):

        self.config = config
        self.index_name = index_name
        self.docs = docs
        self.auto_refresh = auto_refresh
        self.n_batch = batch_size
        self.n_docs = len(docs)
        self.n_chunks = max(1, self.n_docs // self.n_batch)
        self.device = device
        self.process_id = process_id
    
    def process(self):  

        # hf tokenizers setting
        os.environ['TOKENIZERS_PARALLELISM'] = 'false'

        logger.info('starting add documents...')
        
        if self.device.startswith('cpu'):
            logger.info(f"restricting threads to {max_threads_per_process} for process={self.process_id}")
            torch.set_num_threads(max_threads_per_process)
        results = []
        start = time.time()
    
        total_progress_displays = 10
        progress_display_frequency = max(1, self.n_chunks // total_progress_displays)

        for n_processed,_doc in enumerate(np.array_split(self.docs, self.n_chunks)):
            t_chunk_start = time.time()
            percent_done = self._calculate_percent_done(n_processed + 1, self.n_chunks)
            
            if n_processed % progress_display_frequency == 0:
                logger.info(f'process={self.process_id} completed={percent_done}/100% on device={self.device}')
            
            results.append(neural_search.add_documents(config=self.config, index_name=self.index_name, docs=_doc,
                            auto_refresh=self.auto_refresh))
            t_chunk_end = time.time()

            time_left = round((self.n_chunks - (n_processed + 1))*(t_chunk_end - t_chunk_start), 0)
            if n_processed % progress_display_frequency == 0:
                logger.info(f'estimated time left for process {self.process_id} is {time_left} seconds ')

        end = time.time()
        logger.info(f'took {end - start} sec for {self.n_docs} documents')
        return results

    @staticmethod
    def _calculate_percent_done(current_step, total_steps, rounding=0):
        percent = current_step/max(1e-9,total_steps)
        return round(100*percent, rounding)
        
def _run_chunker(chunker: IndexChunk):
    """helper function to run the multiprocess by activating the chunker

    Args:
        chunker (IndexChunk): _description_

    Returns:
        _type_: _description_
    """
    res = chunker.process()
    return res

def add_documents_mp(config=None, index_name=None, docs=None, 
                     auto_refresh=None, batch_size=50):
    """add documents using parallel processing using ray

    Args:
        documents (_type_): _description_
        config (_type_, optional): _description_. Defaults to None.
        index_name (_type_, optional): _description_. Defaults to None.
        auto_refresh (_type_, optional): _description_. Defaults to None.

    Returns:
        _type_: _description_
    """
    n_documents = len(docs)

    logger.info(f"found {n_documents} documents")

    n_processes = get_processes(config.indexing_device)
    if n_documents < n_processes:
        n_processes = max(1, n_documents)
    logger.info(f"using {n_processes} processes")

    # get the device ids for each process based on the process count and available devices
    device_ids = get_device_ids(n_processes, config.indexing_device)
    
    start  = time.time()

    # we create the index if it does not exist
    logger.info("checking index exists and creating if not...")
    try:
        index_info = backend.get_index_info(config=config, index_name=index_name)
    except MarqoApiError as s:
        if s.status_code == 404 and "index_not_found_exception" in str(s):
            neural_search.create_vector_index(config=config, index_name=index_name)
            index_info = backend.get_index_info(config=config, index_name=index_name)
        else:
            raise s

    chunkers = [IndexChunk(config=config, index_name=index_name, docs=_docs,
                                auto_refresh=auto_refresh, batch_size=batch_size, process_id=p_id, device=device_ids[p_id]) 
                                for p_id,_docs in enumerate(np.array_split(docs, n_processes))]
    logger.info('Performing parallel now...')
    with mp.Pool(n_processes) as pool:
        results = pool.map(_run_chunker, chunkers)
    end = time.time()
    logger.info(f"finished indexing all documents. took {end - start} seconds to index {n_documents} documents")
    return results





from typing import List
import functools
import json
from marqo.neural_search import enums
from typing import List, Optional, Union, Callable, Iterable, Sequence, Dict
import copy


def dicts_to_jsonl(dicts: List[dict]) -> str:
    """Turns a list of dicts into a JSONL string"""
    return functools.reduce(
        lambda x, y: "{}\n{}".format(x, json.dumps(y)),
        dicts, ""
    ) + "\n"


def generate_vector_name(field_name: str) -> str:
    """Generates the name of the vector based on the field name"""
    return F"{enums.NeuralField.vector_prefix}{field_name}"


def truncate_dict_vectors(doc: Union[dict, List], new_length: int = 5) -> Union[List, Dict]:
    """Creates a readable version of a dict by truncating identified vectors
    Looks for field names that contains the keyword "vector"
    """
    copied = copy.deepcopy(doc)

    if isinstance(doc, list):
        return [truncate_dict_vectors(d, new_length=new_length)
                if isinstance(d, list) or isinstance(d, dict)
                else copy.deepcopy(d)
                for d in doc]

    for k, v in list(copied.items()):
        if "vector" in k.lower() and isinstance(v, Sequence):
            copied[k] = v[:new_length]
        elif isinstance(v, dict) or isinstance(v, list):
            copied[k] = truncate_dict_vectors(v, new_length=new_length)

    return copied



import functools
import json
import logging
from urllib import parse
from datetime import datetime
from typing import Any, Dict, Generator, List, Optional, Union
from marqo.neural_search import neural_search
from marqo._httprequests import HttpRequests
from marqo.config import Config
from marqo.neural_search.enums import MediaType
from marqo.marqo_logging import logger
from marqo.enums import SearchMethods, Devices
from marqo import errors

# pylint: disable=too-many-public-methods
class Index():
    """
    Indexes routes wrapper.

    Index class gives access to all indexes routes and child routes (inherited).
    https://docs.marqo.com/reference/api/indexes.html
    """

    def __init__(
        self,
        config: Config,
        index_name: str,
        primary_key: Optional[str] = None,
        created_at: Optional[Union[datetime, str]] = None,
        updated_at: Optional[Union[datetime, str]] = None,
    ) -> None:
        """
        Parameters
        ----------
        config:
            Config object containing permission and location of marqo.
        index_name:
            UID of the index on which to perform the index actions.
        primary_key:
            Primary-key of the index.
        """
        self.config = config
        self.http = HttpRequests(config)
        self.index_name = index_name
        self.primary_key = primary_key
        self.created_at = self._maybe_datetime(created_at)
        self.updated_at = self._maybe_datetime(updated_at)

    def delete(self) -> Dict[str, Any]:
        """Delete the index.

        Raises
        ------
        s2SearchApiError
            An error containing details about why marqo can't process your request. marqo error codes are described here: https://docs.marqo.com/errors/#marqo-errors
        """
        return neural_search.delete_index(config=self.config, index_name=self.index_name)

    @staticmethod
    def create(config: Config, index_name: str,
               treat_urls_and_pointers_as_images=False,
               model=None,
               normalize_embeddings=True,
               sentences_per_chunk=2,
               sentence_overlap=0
               ) -> Dict[str, Any]:
        """Create the index.

        Parameters
        ----------
        config:
            config instance
        index_name:
            name of the index.

        Returns
        -------
        task:
            Dictionary containing a task to track the informations about the progress of an asynchronous process.
            https://docs.marqo.com/reference/api/tasks.html#get-one-task

        Raises
        ------
        s2SearchApiError
            An error containing details about why marqo can't process your request. marqo error codes are described here: https://docs.marqo.com/errors/#marqo-errors
        """
        return neural_search.create_vector_index(
            config=config, index_name=index_name, media_type=MediaType.default, neural_settings={
                "index_defaults": {
                    "treat_urls_and_pointers_as_images": treat_urls_and_pointers_as_images,
                    "model": model,
                    "normalize_embeddings": normalize_embeddings,
                    "text_preprocessing": {
                        "split_overlap": sentence_overlap,
                        "split_length": sentences_per_chunk,
                        "split_method": "sentence"
                    }
                }
            }
        )

    def refresh(self):
        """refreshes the index"""
        return self.http.post(path=F"{self.index_name}/_refresh")

    def search(self, q: str, searchable_attributes: Optional[List[str]]=None,
               limit: int=10, search_method: Union[SearchMethods.NEURAL, str] = SearchMethods.NEURAL,
               highlights=True
               ) -> Dict[str, Any]:
        """Search in the index.

        Parameters
        ----------
        query:
            String containing the searched word(s)
        opt_params (optional):
            Dictionary containing optional query parameters
            https://docs.marqo.com/reference/api/search.html#search-in-an-index
        searchable_attributes: a subset of attributes to search through

        Returns
        -------
        results:
            Dictionary with hits, offset, limit, processingTime and initial query

        Raises
        ------
        s2SearchApiError
            An error containing details about why marqo can't process your request. marqo error codes are described here: https://docs.marqo.com/errors/#marqo-errors
        """
        return neural_search.search(
            config=self.config, index_name=self.index_name, text=q, return_doc_ids=True,
            searchable_attributes=searchable_attributes, search_method=search_method, result_count=limit,
            highlights=highlights
        )

    def get_document(self, document_id: Union[str, int]) -> Dict[str, Any]:
        """Get one document with given document identifier.

        Parameters
        ----------
        document_id:
            Unique identifier of the document.

        Returns
        -------
        document:
            Dictionary containing the documents information.

        Raises
        ------
        s2SearchApiError
            An error containing details about why marqo can't process your request. marqo error codes are described here: https://docs.marqo.com/errors/#marqo-errors
        """
        return neural_search.get_document_by_id(
            config=self.config, index_name=self.index_name, document_id=document_id)

    def add_documents(
        self,
        documents: List[Dict[str, Any]],
        auto_refresh=True,
        batch_size: int = None,
        use_parallel: bool = False,
    ) -> Union[Dict[str, Any], List[Dict[str, Any]]]:
        """Add documents to the index.

        Parameters
        ----------
        documents:
            List of documents. Each document should be a dictionary.

        auto_refresh:
            Automatically refresh the index. If you are making lots of requests, it is advised to turn this to
            false to increase performance.

        apply_batching: if True, the documents will be batched into sizes of 'batch_size'

        batch_size: if batch_documents is True, documents will be indexed into batches of this size

        Returns
        -------
        We need to translate into this endpoint: https://www.elastic.co/guide/en/elasticsearch/reference/current/docs-bulk.html

        See Jesse's code for JSONL: https://github.com/S2Search/NeuralSearchPOC/blob/main/data/csv_to_json.py#L152
        Raises
        ------
        s2SearchApiError
            An error containing details about why marqo can't process your request. marqo error codes are described here: https://docs.marqo.com/errors/#marqo-errors
        """

        if batch_size is None:
            return neural_search.add_documents(
                config=self.config, index_name=self.index_name, docs=documents, auto_refresh=auto_refresh)
        elif use_parallel:
            from marqo import parallel
            return parallel.add_documents_mp(config=self.config, index_name=self.index_name, docs=documents, auto_refresh=auto_refresh, batch_size=batch_size)
        else:
            if batch_size <= 0:
                raise errors.MarqoError("Batch size can't be less than 1!")
            return self._batch_request(dataset=documents,batch_size=batch_size, verbose=False)

    def delete_documents(self, ids: List[str], auto_refresh=True) -> Dict[str, int]:
        """Delete multiple documents from the index.

        Parameters
        ----------
        ids:
            List of unique identifiers of documents.
        auto_refresh:
            if true refreshes the index
        Returns
        -------
        task:
            Dictionary containing a task to track the informations about the progress of an asynchronous process.
            https://docs.marqo.com/reference/api/tasks.html#get-one-task

        Raises
        ------
        s2SearchApiError
            An error containing details about why marqo can't process your request. marqo error codes are described here: https://docs.marqo.com/errors/#marqo-errors
        """
        return neural_search.delete_documents(
            config=self.config, index_name=self.index_name, doc_ids=ids,
            auto_refresh=auto_refresh)

    def get_stats(self) -> Dict[str, Any]:
        """Get stats of the index"""
        return neural_search.get_stats(config=self.config, index_name=self.index_name)

    @staticmethod
    def _maybe_datetime(the_date: Optional[Union[datetime, str]]) -> Optional[datetime]:
        """This should handle incoming timestamps from S2Search, including
         parsing if necessary."""
        if the_date is None or not the_date:
            return None

        if isinstance(the_date, datetime):
            return the_date
        elif isinstance(the_date, str):
            parsed_date = datetime.strptime(the_date, "%Y-%m-%dT%H:%M:%S.%f")
            return parsed_date

    def _batch_request(self, dataset, batch_size=100, verbose=True) -> List[Dict[str, Any]]:
        """Batch by the number of documents"""
        logger.info(f"starting batch ingestion in sizes of {batch_size}")

        deeper = ((doc, i, batch_size) for i, doc in enumerate(dataset))

        def batch_requests(gathered, doc_tuple):
            doc, i, the_batch_size = doc_tuple
            if i % the_batch_size == 0:
                gathered.append([doc,])
            else:
                gathered[-1].append(doc)
            return gathered

        batched = functools.reduce(lambda x, y: batch_requests(x, y), deeper, [])

        def verbosely_add_docs(i, docs):
            t0 = datetime.now()
            res = neural_search.add_documents(
                config=self.config, index_name=self.index_name,
                docs=docs, auto_refresh=False)
            total_batch_time = datetime.now() - t0
            num_docs = len(docs)

            logger.info(f"    batch {i}: ingested {num_docs} docs. Time taken: {total_batch_time}. "
                        f"Average timer per doc {total_batch_time/num_docs}")
            if verbose:
                logger.info(f"        results from indexing batch {i}: {res}")
            return res

        results = [verbosely_add_docs(i, docs) for i, docs in enumerate(batched)]
        logger.info('completed batch ingestion.')
        return results

######################################## Class functionality not included in this version#################

    # def add_documents_json(
    #     self,
    #     str_documents: str,
    #     primary_key: Optional[str] = None,
    # ) -> Dict[str, Any]:

    # def add_documents_raw(
    #     self,
    #     str_documents: str,
    #     primary_key: Optional[str] = None,
    #     content_type: Optional[str] = None,
    # ) -> Dict[str, Any]:

    # def add_documents_from_csv_file(
    #         self,
    #         str_documents: str,
    #         primary_key: Optional[str] = None,
    # ) -> Dict[str, Any]:
    #     pass

    # def update_documents(
    #     self,
    #     documents: List[Dict[str, Any]],
    #     primary_key: Optional[str] = None
    # ) -> Dict[str, Any]:
    #     """Update documents in the index.
    #     This can be done using regular add_documents
    #     """

    # def update(self, primary_key: str) -> Dict[str, Any]:
    #     """Update the index primary-key."""

    # def fetch_info(self) -> 'Index':
    #     """Fetch the info of the index. """

    # def get_primary_key(self) -> Optional[str]:
    #     """Get the primary key.
    #     Meilisearch specific. In Meilisearch, users decide which field acts as the ID
    #     """

    # def get_tasks(self) -> Dict[str, List[Dict[str, Any]]]:
    #     """Get all tasks of a specific index from the last one"""

    # def get_task(self, uid: int) -> Dict[str, Any]:
    #     """Get one task through the route of a specific index. """

    # def wait_for_task(
    #     self, uid: int,
    #     timeout_in_ms: int = 5000,
    #     interval_in_ms: int = 50,
    # ) -> Dict[str, Any]:
    #     """Wait until marqo processes a task until it fails or succeeds. """



    # def get_documents(self, parameters: Optional[Dict[str, Any]] = None) -> List[Dict[str, Any]]:
    #     """Get a set of documents from the index.
    #     Will be implemented in near future, probably...
    #       https://docs.meilisearch.com/reference/api/documents.html#get-documents
    #     """

    # def add_documents_in_batches(
    #     self,
    #     documents: List[Dict[str, Any]],
    #     batch_size: int = 1000,
    #     primary_key: Optional[str] = None,
    # ) -> List[Dict[str, Any]]:
    #     """Add documents to the index in batches.
    #     Will probably be implemented later. Seems useful.
    #     """

    # def add_documents_csv(
    #     self,
    #     str_documents: str,
    #     primary_key: Optional[str] = None,
    # ) -> Dict[str, Any]:
    #     """Add string documents from a CSV file to the index."""

    # def add_documents_ndjson(
    #     self,
    #     str_documents: str,
    #     primary_key: Optional[str] = None,
    # ) -> Dict[str, Any]:
    #     """Add string documents from a NDJSON file to the index."""

    # def update_documents_in_batches(
    #     self,
    #     documents: List[Dict[str, Any]],
    #     batch_size: int = 1000,
    #     primary_key: Optional[str] = None
    # ) -> List[Dict[str, Any]]:
    #     """Update documents to the index in batches."""

    # def delete_all_documents(self) -> Dict[str, int]:
    #     """Delete all documents from the index. """

    # # GENERAL SETTINGS ROUTES

    # def get_settings(self) -> Dict[str, Any]:
    #     """Get settings of the index."""
    #
    # def update_settings(self, body: Dict[str, Any]) -> Dict[str, Any]:
    #     """Update settings of the index. """
    #
    # def reset_settings(self) -> Dict[str, Any]:
    #     """Reset settings of the index to default values."""

    # # RANKING RULES SUB-ROUTES

    # def get_ranking_rules(self) -> List[str]:
    #     """Get ranking rules of the index."""
    #
    # def update_ranking_rules(self, body: List[str]) -> Dict[str, Any]:
    #     """Update ranking rules of the index."""
    #
    # def reset_ranking_rules(self) -> Dict[str, Any]:
    #     """Reset ranking rules of the index to default values."""

    # # DISTINCT ATTRIBUTE SUB-ROUTES

    # def get_distinct_attribute(self) -> Optional[str]:
    #     """Get distinct attribute of the index. """
    #
    # def update_distinct_attribute(self, body: Dict[str, Any]) -> Dict[str, Any]:
    #     """Update distinct attribute of the index."""
    #
    # def reset_distinct_attribute(self) -> Dict[str, Any]:
    #     """Reset distinct attribute of the index to default values."""

    # # SEARCHABLE ATTRIBUTES SUB-ROUTES

    # def get_searchable_attributes(self) -> List[str]:
    #     """Get searchable attributes of the index."""
    #
    # def update_searchable_attributes(self, body: List[str]) -> Dict[str, Any]:
    #     """Update searchable attributes of the index."""
    #
    # def reset_searchable_attributes(self) -> Dict[str, Any]:
    #     """Reset searchable attributes of the index to default values."""

    # # DISPLAYED ATTRIBUTES SUB-ROUTES

    # def get_displayed_attributes(self) -> List[str]:
    #     """Get displayed attributes of the index."""
    #
    # def update_displayed_attributes(self, body: List[str]) -> Dict[str, Any]:
    #     """Update displayed attributes of the index. """
    #
    # def reset_displayed_attributes(self) -> Dict[str, Any]:
    #     """Reset displayed attributes of the index to default values. """

    # # STOP WORDS SUB-ROUTES

    # def get_stop_words(self) -> List[str]:
    #     """Get stop words of the index."""
    #
    # def update_stop_words(self, body: List[str]) -> Dict[str, Any]:
    #     """Update stop words of the index."""
    #
    # def reset_stop_words(self) -> Dict[str, Any]:
    #     """Reset stop words of the index to default values."""

    # # SYNONYMS SUB-ROUTES

    # def get_synonyms(self) -> Dict[str, List[str]]:
    #     """Get synonyms of the index."""
    #
    # def update_synonyms(self, body: Dict[str, List[str]]) -> Dict[str, Any]:
    #     """Update synonyms of the index."""
    #
    # def reset_synonyms(self) -> Dict[str, Any]:
    #     """Reset synonyms of the index to default values."""

    # # FILTERABLE ATTRIBUTES SUB-ROUTES

    # def get_filterable_attributes(self) -> List[str]:
    #     """Get filterable attributes of the index."""
    #
    # def update_filterable_attributes(self, body: List[str]) -> Dict[str, Any]:
    #     """Update filterable attributes of the index. """
    #
    # def reset_filterable_attributes(self) -> Dict[str, Any]:
    #     """Reset filterable attributes of the index to default values."""

    # # SORTABLE ATTRIBUTES SUB-ROUTES

    # def get_sortable_attributes(self) -> List[str]:
    #     """Get sortable attributes of the index."""
    #
    # def update_sortable_attributes(self, body: List[str]) -> Dict[str, Any]:
    #     """Update sortable attributes of the index."""
    #
    # def reset_sortable_attributes(self) -> Dict[str, Any]:
    #     """Reset sortable attributes of the index to default values."""

    # # TYPO TOLERANCE SUB-ROUTES

    # def get_typo_tolerance(self) -> Dict[str, Any]:
    #     """Get typo tolerance of the index."""
    #
    # def update_typo_tolerance(self, body: Dict[str, Any]) -> Dict[str, Any]:
    #     """Update typo tolerance of the index"""
    #
    # def reset_typo_tolerance(self) -> Dict[str, Any]:
    #     """Reset typo tolerance of the index to default values."""

    # # HELPERS

    # @staticmethod
    # def _batch(
    #         documents: List[Dict[str, Any]], batch_size: int
    # ) -> Generator[List[Dict[str, Any]], None, None]:
    #     total_len = len(documents)
    #     for i in range(0, total_len, batch_size):
    #         yield documents[i: i + batch_size]

    # def __settings_url_for(self, sub_route: str) -> str:
    #     return f'{self.config.paths.index}/{self.index_name}/{self.config.paths.setting}/{sub_route}'



import json

from requests import Response

class MarqoError(Exception):
    """Generic class for Marqo error handling"""

    def __init__(self, message: str) -> None:
        self.message = message
        super().__init__(self.message)

    def __str__(self) -> str:
        return f'MarqoError. Error message: {self.message}'


class MarqoApiError(MarqoError):
    """Error sent by Marqo API"""

    def __init__(self, error: str, request: Response) -> None:
        self.status_code = request.status_code
        self.code = None
        self.link = None
        self.type = None

        if request.text:
            json_data = json.loads(request.text)
            self.message = json_data
            self.code = json_data.get('status')
            self.link = ''
            self.type = ''
            if 'error' in json_data and 'root_cause' in json_data["error"]\
                    and len(json_data.get('error').get('root_cause')) > 0:
                self.type = json_data.get('error').get('root_cause')[0].get('type')
        else:
            self.message = error
        super().__init__(self.message)

    def __str__(self) -> str:
        if self.code and self.link:
            return f'MarqoApiError. Error code: {self.code}. Error message: {self.message} Error documentation: {self.link} Error type: {self.type}'

        return f'MarqoApiError. {self.message}'


class MarqoNonNeuralIndexError(MarqoError):
    """Error trying to use a non-neural index like a neural one"""

    def __str__(self) -> str:
        return f'MarqoCommunicationError, {self.message}'


class MarqoCommunicationError(MarqoError):
    """Error when connecting to Marqo"""

    def __str__(self) -> str:
        return f'MarqoCommunicationError, {self.message}'


class MarqoTimeoutError(MarqoError):
    """Error when Marqo operation takes longer than expected"""

    def __str__(self) -> str:
        return f'MarqoTimeoutError, {self.message}'


# implements a 'model' that returns a random vector, irrespective of 
# input. does not require a model and is completely random
# used for testing purposes
from marqo.s2_inference.sbert_utils import Model
from marqo.s2_inference.types import Union, FloatTensor, List, ndarray

import numpy as np
import hashlib

def sentence_to_hash(sentence):
    return int(hashlib.sha256(sentence.encode('utf-8')).hexdigest(), 16) % 10**8

class Random(Model):

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

    def load(self) -> None:
        self.model = None

    @staticmethod
    def _get_sentence_hash(sentence):
        return sentence_to_hash(sentence)

    def _get_sentences_hash(self, sentences):
        hashes = 0
        i= 0
        for i,sentence in enumerate(sentences):
            hashes += self._get_sentence_hash(sentence)
        return hashes // (i + 1)

    def encode(self, sentence: Union[str, List[str]], normalize: bool = True, **kwargs) -> Union[FloatTensor, ndarray]:

        if self.embedding_dimension is None or self.embedding_dimension == 0:
            raise ValueError('invalid embedding dimension size. check the model registry is correct')

        if isinstance(sentence, str):
            # same strings generate same random vector (I think)
            seed = self._get_sentence_hash(sentence)
            np.random.seed(seed)

            return np.random.rand(1, self.embedding_dimension)
        else:
            if len(sentence) == 0:
                raise ValueError('recevied empty sentence')

            seed = self._get_sentences_hash(sentence)
            np.random.seed(seed)

            return np.random.rand(len(sentence), self.embedding_dimension)



class MediaType:
    text = 'text'
    image = 'image'
    default = 'text'

class Normalization:
    normalize = True 



class MlModel:
    bert = "hf/all_datasets_v4_MiniLM-L6"
    clip = "ViT-L/14"

class SearchMethod:
    # BM25/ TF-IDF:
    LEXICAL = "LEXICAL"
    # chunk_embeddings
    NEURAL = "NEURAL"

class NeuralField:
    """Neural Search fields
    Protected document field names"""
    field_name = "__field_name"
    field_content = "__field_content"
    doc_chunk_relation = "__doc_chunk_relation"
    chunk_ids = "__chunk_ids"
    # the prefix will have the customer's field name appended to the end of it
    vector_prefix = "__vector_"
    chunks = "__chunks"
    output_highlights = "_highlights"
    output_score = "_score"

class NeuralSettingsField:
    neural_settings = "neural_settings"
    index_defaults = "index_defaults"
    treat_urls_and_pointers_as_images = "treat_urls_and_pointers_as_images"
    model = "model"
    normalize_embeddings = "normalize_embeddings"

    text_preprocessing = "text_preprocessing"
    split_length = "split_length"
    split_overlap = "split_overlap"
    split_method = "split_method"


class SplitMethod:
    # consider moving this enum into processing
    sentence = "sentence"




from statistics import mode
# from typing import List, Dict, Tuple, Iterable, Type, Union, Callable, Optional
# from torch import FloatTensor

import torch
import os
import onnxruntime
from transformers import AutoModel, AutoTokenizer
from pathlib import Path

from marqo.s2_inference.types import *
from marqo.s2_inference.logger import get_logger
logger = get_logger(__name__)

class ModelCache:

    onnx_cache_path = os.environ.get('ONNX_SAVE_PATH', './cache/models_onnx/')
    torch_cache_path = os.getenv('SENTENCE_TRANSFORMERS_HOME', './cache/models/')

class BaseTransformerModels:

    names = ('albert-base-v1', 'albert-base-v2', 'albert-large-v1', 'albert-large-v2', 'albert-xlarge-v1', 'albert-xlarge-v2', 'albert-xxlarge-v1', 'albert-xxlarge-v2', 'bert-base-cased-finetuned-mrpc', 'bert-base-cased', 'bert-base-chinese', 'bert-base-german-cased', 'bert-base-german-dbmdz-cased', 'bert-base-german-dbmdz-uncased', 'bert-base-multilingual-cased', 'bert-base-multilingual-uncased', 'bert-base-uncased', 'bert-large-cased-whole-word-masking-finetuned-squad', 'bert-large-cased-whole-word-masking', 'bert-large-cased', 'bert-large-uncased-whole-word-masking-finetuned-squad', 'bert-large-uncased-whole-word-masking', 'bert-large-uncased', 'camembert-base', 'ctrl', 'distilbert-base-cased-distilled-squad', 'distilbert-base-cased', 'distilbert-base-german-cased', 'distilbert-base-multilingual-cased', 'distilbert-base-uncased-distilled-squad',
                                    'distilbert-base-uncased-finetuned-sst-2-english', 'distilbert-base-uncased', 'distilgpt2', 'distilroberta-base', 'gpt2-large', 'gpt2-medium', 'gpt2-xl', 'gpt2', 'openai-gpt', 'roberta-base-openai-detector', 'roberta-base', 'roberta-large-mnli', 'roberta-large-openai-detector', 'roberta-large', 't5-11b', 't5-3b', 't5-base', 't5-large', 't5-small', 'transfo-xl-wt103', 'xlm-clm-ende-1024', 'xlm-clm-enfr-1024', 'xlm-mlm-100-1280', 'xlm-mlm-17-1280', 'xlm-mlm-en-2048', 'xlm-mlm-ende-1024', 'xlm-mlm-enfr-1024', 'xlm-mlm-enro-1024', 'xlm-mlm-tlm-xnli15-1024', 'xlm-mlm-xnli15-1024', 'xlm-roberta-base', 'xlm-roberta-large-finetuned-conll02-dutch', 'xlm-roberta-large-finetuned-conll02-spanish', 'xlm-roberta-large-finetuned-conll03-english', 'xlm-roberta-large-finetuned-conll03-german', 'xlm-roberta-large', 'xlnet-base-cased', 'xlnet-large-cased')

class Invalid:

    paths = (None, '', ' ', "", " ", "''", '""')

class Ignore:

    files = ('flax_model.msgpack', 'rust_model.ot', 'tf_model.h5')



class SBERT_ONNX(object):

    """takes and existing huggingface or sbert model and converts to onnx and loads ready for inference on cpu or gpu
    TODO: split the convert and loading/encoding
    """

    def __init__(self, model_name_or_path: Optional[str] = None,
                 device: Optional[str] = None,
                 cache_folder: Optional[str] = None,
                 onnx_folder: Optional[str] = None,
                 onnx_model_name: Optional[str] = None,
                 embedding_dim: Optional[str] = None,
                 enable_overwrite: Optional[bool] = False,
                 max_seq_length: int = 128,
                 lower_case: bool = True,
                 ):
       

        self.device = device
        self.cache_folder = cache_folder
        self.onnx_folder = onnx_folder
        self.onnx_model_name = onnx_model_name
        self.enable_overwrite = enable_overwrite
        self.model_name_or_path = model_name_or_path
        self.max_seq_length = max_seq_length
        self.do_lower_case = lower_case
        self.embedding_dim = embedding_dim

        self.fast_onnxprovider = None
        self.onnxproviders = None
        self.model_path = None
        self.export_model_name = None
        self.model = None
        self.tokenizer = None
        self.session = None

        self._get_paths()
        self._get_onnx_provider()

    def load(self) -> None:
        """this does all the steps to get the onnx model
        """
        self._prepare()
        self._convert_to_onnx()
        self._load_sbert_session()
        logger.info(f"loaded {self.onnx_model_name} succesfully")

    def _get_paths(self) -> None:
        """get the paths of the cache, onnx save path and output model path
        """
        if self.onnx_folder is None:
            self.onnx_folder = ModelCache.onnx_cache_path
            Path(self.onnx_folder).mkdir(parents=True, exist_ok=True)

        if self.cache_folder is None:
            self.cache_folder = ModelCache.torch_cache_path

        if self.onnx_model_name is None:
            self.onnx_model_name = f"{os.path.basename(self.model_name_or_path.replace('/', '_'))}.onnx"
    
        self.export_model_name = os.path.join(self.onnx_folder, f"{self.onnx_model_name}") 

    def _get_onnx_provider(self) -> None:
        """determine where the model should run based on specified device
        """
        self.onnxproviders = onnxruntime.get_available_providers()
        logger.info(f"device:{self.device}")
        if self.device == 'cpu':
            self.fast_onnxprovider = 'CPUExecutionProvider'
        else:
            if 'CUDAExecutionProvider' not in self.onnxproviders:
                self.fast_onnxprovider = 'CPUExecutionProvider'
            else:
                self.fast_onnxprovider = 'CUDAExecutionProvider'

        logger.info(f"onnx_provider:{self.fast_onnxprovider}")

    def _prepare(self) -> None:
        """load the model and put it in eval mode
        """
        self.tokenizer = AutoTokenizer.from_pretrained(
            self.model_name_or_path, do_lower_case=self.do_lower_case)
        self.model = AutoModel.from_pretrained(
            self.model_name_or_path)

        self.model.eval()

    def _load_sbert_session(self) -> None:

        """
        create the onnx session for running the model
        also see 
        https://github.com/microsoft/onnxruntime/tree/master/onnxruntime/python/tools/transformers
        https://github.com/microsoft/onnxruntime/blob/master/onnxruntime/python/tools/transformers/bert_perf_test.py
        """

        sess_options = onnxruntime.SessionOptions()
        # sess_options.graph_optimization_level = onnxruntime.GraphOptimizationLevel.ORT_ENABLE_ALL
        self.session = onnxruntime.InferenceSession(
            self.export_model_name, sess_options, providers=[self.fast_onnxprovider])

    def _convert_to_onnx(self) -> None:
        """converts from pytorch to onnx
        """
        st = ['hello, how are you']
        inputs = self.tokenizer(
            st,
            padding=True,
            truncation=True,
            max_length=self.max_seq_length,
            return_tensors="pt")

        if self.enable_overwrite or not os.path.exists(self.export_model_name):
            with torch.no_grad():
                symbolic_names = {0: 'batch_size', 1: 'max_seq_len'}
                torch.onnx.export(self.model,                                            # model being run
                                  # model input (or a tuple for multiple inputs)
                                  args=tuple(inputs.values()),
                                  # where to save the model (can be a file or file-like object)
                                  f=self.export_model_name,
                                  # the ONNX version to export the model to
                                  opset_version=11,
                                  # whether to execute constant folding for optimization
                                  do_constant_folding=True,
                                  input_names=['input_ids',                         # the model's input names
                                               'attention_mask',
                                               'token_type_ids'],
                                  # the model's output names
                                  output_names=['start', 'end'],
                                  dynamic_axes={'input_ids': symbolic_names,        # variable length axes
                                                'attention_mask': symbolic_names,
                                                'token_type_ids': symbolic_names,
                                                'start': symbolic_names,
                                                'end': symbolic_names})
            logger.info(f"Model exported at: {self.export_model_name}")

            # from onnxruntime.transformers import optimizer
            # optimized_model = optimizer.optimize_model(self.export_model_name, model_type='bert', num_heads=6, hidden_size=768//2)
            # optimized_model.convert_float_to_float16()
            # optimized_model.save_model_to_file(self.export_model_name)

    @staticmethod
    def normalize(outputs: FloatTensor) -> FloatTensor:
        """normalizes vector or matrix to have unit length across rows

        Args:
            outputs (_type_): _description_

        Returns:
            _type_: _description_
        """
        return outputs/outputs.norm(dim=-1, keepdim=True)

    @staticmethod
    def mean_pooling(token_embeddings: FloatTensor, attention_mask: FloatTensor) -> FloatTensor:
        """performs mean average pooling over a matrix with a supplied attention mask

        Args:
            token_embeddings (FloatTensor): _description_
            attention_mask (FloatTensor): _description_

        Returns:
            FloatTensor: _description_
        """
        input_mask_expanded = attention_mask.unsqueeze(-1).expand(token_embeddings.size()).float()

        return torch.sum(token_embeddings * input_mask_expanded, 1) / torch.clamp(input_mask_expanded.sum(1), min=1e-9)

    def encode(self, sentences: Union[str, List[str]], normalize: bool = True, **kwargs) -> FloatTensor:
        """performs the vectorizxation of a sentence or list of sentences

        Args:
            sentences (Union[str, List[str]]): _description_
            normalize (bool, optional): _description_. Defaults to Normalization.normalize.

        Returns:
            FloatTensor: _description_
        """
        if isinstance(sentences, str) or not hasattr(sentences, '__len__'):
            sentences = [sentences]

        # maybbe change to np and remove the .cpu().numpy calss?
        inputs = self.tokenizer(
            sentences,
            padding=True,
            truncation=True,
            max_length=self.max_seq_length,
            return_tensors="pt"
        )
        ort_inputs = {k: v.cpu().numpy() for k, v in inputs.items()}

        ort_outputs = self.session.run(None, ort_inputs)
        result = self.mean_pooling(token_embeddings=torch.FloatTensor(ort_outputs[0]),
                                                        attention_mask=inputs.get('attention_mask'))
        if normalize:
            return self.normalize(result)

        return result


    
        



import pprint
from typing import NamedTuple, Any
from marqo.neural_search import enums


class IndexInfo(NamedTuple):
    """
    model_name: name of the ML model used to encode the data
    properties: keys are different index field names, values
        provide info about the properties
    """
    model_name: str
    properties: dict
    neural_settings: dict

    def get_neural_settings(self) -> dict:
        return self.neural_settings.copy()

    def get_vector_properties(self) -> dict:
        """returns a dict containing only names and properties of vector fields
        Perhaps a better approach is to check if the field's props is actually a vector type,
        plus checks over fieldnames
        """
        return {
            vector_name: vector_props
            for vector_name, vector_props in self.properties[enums.NeuralField.chunks]["properties"].items()
            if vector_name.startswith(enums.NeuralField.vector_prefix)
        }

    def get_text_properties(self):
        """returns a dict containing only names and properties of text fields
        Perhaps a better approach is to check if the text's props is actually a text type,
        plus checks over fieldnames"""
        return {
            text_field: text_props
            for text_field, text_props in self.properties.items()
            if not text_field.startswith(enums.NeuralField.vector_prefix)
                and not text_field in enums.NeuralField.__dict__.values()
        }


from sentence_transformers import SentenceTransformer
import numpy as np
from torch import nn

from marqo.s2_inference.types import *

from marqo.s2_inference.logger import get_logger
logger = get_logger(__name__)


class Model:
    """ generic model wrapper class
    """
    def __init__(self, model_name: str, device: str = 'cpu', batch_size: int = 2048, embedding_dim=None, max_seq_length=None) -> None:

        self.model_name = model_name
        self.device = device
        self.model = None
        self.embedding_dimension = embedding_dim
        self.batch_size = batch_size
        self.max_seq_length = max_seq_length

    def load(self) -> None:
        """ method to load the model
        """
        pass

    def encode(self, sentence: Union[str, List[str]]) -> None:
        """encodes a single sentece or multiple sentences

        Args:
            sentence (str): _description_
        """
        pass


class SBERT(Model):
    """class for SBERT models

    Args:
        Model (_type_): _description_
    """
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

    def load(self) -> None:
        self.model = SentenceTransformer(self.model_name, device=self.device)

        # if one provided, overrite
        if self.max_seq_length is None:
            self.max_seq_length = self.model.max_seq_length
        else:
            self.model.max_seq_length = self.max_seq_length


    def encode(self, sentence: Union[str, List[str]], normalize = True, **kwargs) -> Union[FloatTensor, np.ndarray]:

        if self.model is None:
            self.load()

        if isinstance(sentence, str):
            sentence = [sentence]

        # seemed inconsistent with the normalization, = False, roll own
        embeddings = self.model.encode(sentence, batch_size=self.batch_size, 
                        normalize_embeddings=False, convert_to_tensor=True)

        if normalize:
            embeddings = nn.functional.normalize(embeddings, p=2, dim=1)

        return embeddings

class TEST(Model):
    """class for SBERT models

    Args:
        Model (_type_): _description_
    """
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

    def load(self) -> None:
        self.model = SentenceTransformer(self.model_name, device=self.device)

    def encode(self, sentence: Union[str, List[str]], **kwargs) -> Union[FloatTensor, np.ndarray]:

        if self.model is None:
            self.load()

        if isinstance(sentence, str):
            sentence = [sentence]

        return self.model.encode(sentence, batch_size=self.batch_size, **kwargs)[:, :16]


import copy
import json
import pprint
from typing import Any, Callable, Dict, List, Optional, Union
import requests
from marqo.config import Config
from marqo.errors import (
    MarqoApiError,
    MarqoCommunicationError,
    MarqoTimeoutError,
)
from marqo.version import qualified_version

ALLOWED_OPERATIONS = {requests.delete, requests.get, requests.post, requests.put}

OPERATION_MAPPING = {'delete': requests.delete, 'get': requests.get,
                     'post': requests.post, 'put': requests.put}


class HttpRequests:
    def __init__(self, config: Config) -> None:
        self.config = config
        self.headers = dict()

    def send_request(
        self,
        http_method: Callable,
        path: str,
        body: Optional[Union[Dict[str, Any], List[Dict[str, Any]], List[str], str]] = None,
        content_type: Optional[str] = None,
    ) -> Any:
        to_verify = self.config.cluster_is_remote

        if http_method not in ALLOWED_OPERATIONS:
            raise ValueError("{} not an allowed operation {}".format(http_method, ALLOWED_OPERATIONS))

        req_headers = copy.deepcopy(self.headers)

        if content_type is not None and content_type:
            req_headers['Content-Type'] = content_type

        try:
            request_path = self.config.url + '/' + path
            if isinstance(body, bytes):
                response = http_method(
                    request_path,
                    timeout=self.config.timeout,
                    headers=req_headers,
                    data=body,
                    verify=to_verify
                )
            elif isinstance(body, str):
                response = http_method(
                    request_path,
                    timeout=self.config.timeout,
                    headers=req_headers,
                    data=body,
                    verify=to_verify
                )
            else:
                response = http_method(
                    request_path,
                    timeout=self.config.timeout,
                    headers=req_headers,
                    data=json.dumps(body) if body else None,
                    verify=to_verify
                )
            return self.__validate(response)

        except requests.exceptions.Timeout as err:
            raise MarqoTimeoutError(str(err)) from err
        except requests.exceptions.ConnectionError as err:
            raise MarqoCommunicationError(str(err)) from err

    def get(
        self, path: str,
        body: Optional[Union[Dict[str, Any], List[Dict[str, Any]], List[str], str]] = None,
    ) -> Any:
        content_type = None
        if body is not None:
            content_type = 'application/json'
        return self.send_request(requests.get, path=path, body=body, content_type=content_type)

    def post(
        self,
        path: str,
        body: Optional[Union[Dict[str, Any], List[Dict[str, Any]], List[str], str]] = None,
        content_type: Optional[str] = 'application/json',
    ) -> Any:
        return self.send_request(requests.post, path, body, content_type)

    def put(
        self,
        path: str,
        body: Optional[Union[Dict[str, Any], List[Dict[str, Any]], List[str], str]] = None,
        content_type: Optional[str] = None,
    ) -> Any:
        if body is not None:
            content_type = 'application/json'
        return self.send_request(requests.put, path, body, content_type)

    def delete(
        self,
        path: str,
        body: Optional[Union[Dict[str, Any], List[Dict[str, Any]], List[str]]] = None,
    ) -> Any:
        return self.send_request(requests.delete, path, body)

    @staticmethod
    def __to_json(
        request: requests.Response
    ) -> Any:
        if request.content == b'':
            return request
        return request.json()

    @staticmethod
    def __validate(
        request: requests.Response
    ) -> Any:
        try:
            request.raise_for_status()
            return HttpRequests.__to_json(request)
        except requests.exceptions.HTTPError as err:
            raise MarqoApiError(str(err), request) from err


# from torch import FloatTensor
# from typing import Any, Dict, List, Optional, Union
import os
import validators
import requests
import numpy as np
import clip
import torch
from PIL import Image

from marqo.s2_inference.types import *
from marqo.s2_inference.logger import get_logger
logger = get_logger(__name__)


def get_allowed_image_types():
    return set(('.jpg', '.png', '.bmp', '.jpeg'))


def format_and_load_CLIP_images(images: List[Union[str, ndarray, ImageType]]) -> List[ImageType]:
    """takes in a list of strings, arrays or urls and either loads and/or converts to PIL
        for the clip model

    Args:
        images (List[Union[str, np.ndarray, ImageType]]): list of file locations or arrays (can be mixed)

    Raises:
        TypeError: _description_

    Returns:
        List[ImageType]: list of PIL images
    """
    if not isinstance(images, list):
        raise TypeError(f"expected list but received {type(images)}")

    results = []
    for image in images:
        results.append(format_and_load_CLIP_image(image))
    
    return results

def format_and_load_CLIP_image(image: Union[str, ndarray, ImageType]) -> ImageType:
    """standardizes the input to be a PIL image

    Args:
        image (Union[str, np.ndarray, ImageType]): can be a local file, url or array

    Raises:
        ValueError: _description_
        TypeError: _description_

    Returns:
        ImageType: PIL image
    """
    # check for the input type
    if isinstance(image, str):
        if os.path.isfile(image):
            img = Image.open(image)
        elif validators.url(image):
            img = Image.open(requests.get(image, stream=True).raw)
        else:
            raise ValueError(f"input str of {image} is not a local file or a valid url")
    elif isinstance(image, np.ndarray):
        img = Image.fromarray(image.astype('uint8'), 'RGB')

    elif isinstance(image, ImageType):
        img = image
    else:
        raise TypeError(f"input of type {type(image)} did not match allowed types of str, np.ndarray, ImageType")

    return img

def _is_image(inputs: Union[str, List[Union[str, ImageType, ndarray]]]) -> bool:
    # some logic to determine if something is an image or not
    # assume the batch is the same type
    # maybe we use something like this https://github.com/ahupp/python-magic
    
    _allowed = get_allowed_image_types()

    # we assume the batch is this way if a list
    # otherwise apply over each element
    if isinstance(inputs, list):

        if len(inputs) == 0:
            raise TypeError("received empty list, expected at least one element.")

        thing = inputs[0]
    else:
        thing = inputs
    
    # if it is a string, determine if it is a local file or url
    if isinstance(thing, str):
        name, extension = os.path.splitext(thing.lower())
        
        # if it has the correct extension, asssume yes
        if extension in _allowed:
            return True
        
        # if it is a local file without extension, then raise an error
        if os.path.isfile(thing):
            # we could also read the first part of the file and infer
            raise TypeError(f"local file [{thing}] extension {extension} does not match allowed file types of {_allowed}")
        else:
            # if it is not a local file and does not have an extension
            # check if url
            if validators.url(thing):
                return True
            else:
                False
                # raise ValueError(f"{thing} cannot be identified as a local file, url or image")

    # if it is an array, then it is an image
    elif isinstance(thing, (ImageType, ndarray)):
        return True
    else:
        raise TypeError(f"expected type Image or str for inputs but received type {type(thing)}")

class CLIP:
    
    """
    conveniance class wrapper to make clip work easily for both text and image encoding
    """

    def __init__(self, model_type: str = "ViT-B/32", device: str = 'cpu',  embedding_dim: int = None,
                            truncate: bool = True, **kwargs) -> None:

        self.model_type = model_type
        self.device = device
        self.model = None
        self.tokenizer = None
        self.processor = None
        self.embedding_dimension = embedding_dim
        self.truncate = truncate

    def load(self) -> None:

        # https://github.com/openai/CLIP/issues/30
        self.model, self.preprocess = clip.load(self.model_type, device='cpu', jit=False)
        self.model = self.model.to(self.device)
        self.tokenizer = clip.tokenize
        self.model.eval()
    
    def _convert_output(self, output):

        if self.device == 'cpu':
            return output.numpy()
        elif self.device.startswith('cuda'):
            return output.cpu().numpy()

    @staticmethod
    def normalize(outputs):
        return outputs.norm(dim=-1, keepdim=True)

    def encode_text(self, sentence: Union[str, List[str]], normalize = True) -> FloatTensor:
        
        if self.model is None:
            self.load()
        
        text = self.tokenizer(sentence, truncate=self.truncate).to(self.device)

        with torch.no_grad():
            outputs =  self.model.encode_text(text)

        if normalize:
            _shape_before = outputs.shape
            outputs /= self.normalize(outputs)
            assert outputs.shape == _shape_before

        return self._convert_output(outputs)

    def encode_image(self, images: Union[str, ImageType, List[Union[str, ImageType]]], 
                        normalize = True) -> FloatTensor:
        
        if self.model is None:
            self.load()

        # default to batch encoding
        if isinstance(images, list):
            image_input = format_and_load_CLIP_images(images)
        else:
            image_input = [format_and_load_CLIP_image(images)]

        self.image_input_processed = torch.stack([self.preprocess(_img).to(self.device) for _img in image_input])
    
        with torch.no_grad():
            outputs = self.model.encode_image(self.image_input_processed)

        if normalize:
            _shape_before = outputs.shape
            outputs /= self.normalize(outputs)
            assert outputs.shape == _shape_before
        return self._convert_output(outputs)

    def encode(self, inputs: Union[str, ImageType, List[Union[str, ImageType]]], 
                                default: str = 'text', normalize = True, **kwargs) -> FloatTensor:

        infer = kwargs.pop('infer', True)

        if infer and _is_image(inputs):
            is_image = True
        else:
            is_image = False
            if default == 'text':
                is_image = False
            elif default == 'image':
                is_image = True
            else:
                raise ValueError(f"expected default='image' or default='text' but received {default}")

        if is_image:
            logger.debug('image')
            return self.encode_image(inputs, normalize=normalize)
        else:
            logger.debug('text')
            return self.encode_text(inputs, normalize=normalize)
        



"""Communication with Neural Search's persistence and search layer (OpenSearch)"""
import json
import pprint
from marqo.neural_search.models.index_info import IndexInfo
# client-specific modules - we may want to replace these:
from marqo._httprequests import HttpRequests
from marqo.config import Config
from marqo.errors import MarqoError
from marqo.neural_search import validation, constants, enums
from marqo.neural_search import utils
from marqo import errors
#
from typing import Iterable, List, Union, Optional
from marqo.neural_search.index_meta_cache import get_cache


def get_index_info(config: Config, index_name: str) -> IndexInfo:
    """Gets useful information about the index. Also updates the IndexInfo cache

    Args:
        config:
        index_name:

    Returns:
        IndexInfo of the index

    Raises:
        MarqoError, if the index's mapping doesn't conform to a Neural Search index

    """
    res = HttpRequests(config).get(path=F"{index_name}/_mapping")

    if not (index_name in res and "mappings" in res[index_name]
            and "_meta" in res[index_name]["mappings"]):
        raise errors.MarqoNonNeuralIndexError(
            f"Error retrieving index info for index {index_name}")

    if "model" in res[index_name]["mappings"]["_meta"]:
        model_name = res[index_name]["mappings"]["_meta"]["model"]
    else:
        raise errors.MarqoNonNeuralIndexError(
            "get_index_info: couldn't identify embedding model name "
            F"in index mappings! Mapping: {res}")

    if "neural_settings" in res[index_name]["mappings"]["_meta"]:
        neural_settings = res[index_name]["mappings"]["_meta"]["neural_settings"]
    else:
        raise errors.MarqoNonNeuralIndexError(
            "get_index_info: couldn't identify neural_settings "
            F"in index mappings! Mapping: {res}")

    index_properties = res[index_name]["mappings"]["properties"]

    index_info = IndexInfo(model_name=model_name, properties=index_properties,
                           neural_settings=neural_settings)
    get_cache()[index_name] = index_info
    return index_info


def add_customer_field_properties(config: Config, index_name: str, customer_field_names: Iterable[str], model_properties: dict):
    """Adds new customer fields to index mapping.

    Args:
        config:
        index_name:
        customer_field_names: the new fieldnames the customers have made
        model_properties: properties of the machine learning model

    Returns:
        HTTP Response
    """
    body = {
        "properties": {
            enums.NeuralField.chunks: {
                "type": "nested",
                "properties": {
                    validation.validate_vector_name(
                        utils.generate_vector_name(field_name)): {
                        "type": "knn_vector",
                        "dimension": model_properties["dimensions"],
                        "method": {
                            "name": "hnsw",
                            "space_type": "innerproduct",
                            "engine": "nmslib",
                            "parameters": {
                                "ef_construction": 128,
                                "m": 24
                            }
                        }
                    } for field_name in customer_field_names
                }
            }
        }
    }

    existing_info = get_cache()[index_name]
    new_index_properties = existing_info.properties.copy()

    for field_name in customer_field_names:
        body["properties"][validation.validate_field_name(field_name)] = {
            "type": "text"
        }
        new_index_properties[validation.validate_field_name(field_name)] = {
            "type": "text"
        }

    merged_chunk_properties = {
        **existing_info.properties[enums.NeuralField.chunks]["properties"],
        **body["properties"][enums.NeuralField.chunks]["properties"]
    }
    new_index_properties[enums.NeuralField.chunks]["properties"] = merged_chunk_properties
    get_cache()[index_name] = IndexInfo(
        model_name=existing_info.model_name,
        properties=new_index_properties,
        neural_settings=existing_info.neural_settings.copy()
    )
    mapping_res = HttpRequests(config).put(path=F"{index_name}/_mapping", body=json.dumps(body))
    return mapping_res


def get_cluster_indices(config: Config):
    """Gets the name of all indices"""
    res = HttpRequests(config).get(path="_aliases")
    indices = set(res.keys())
    relevant_indices = indices - constants.INDEX_NAMES_TO_IGNORE
    return relevant_indices



from typing import Any, Dict, List, Optional, Union
from types import FunctionType

from functools import partial
from more_itertools import windowed

from nltk.tokenize import sent_tokenize, word_tokenize
import nltk


def _splitting_functions(split_by: str, language: str='english') -> FunctionType:
    """_summary_
    selects a text splitting function based on the method provided by 'split_by'
    Args:
        split_by (str): method to split the text by, 'character', 'word', 'sentence', 'passage'
                        if not one of those allows for custom characters to split on
        language (str, optional): _description_. Defaults to 'english'.

    Raises:
        TypeError: _description_

    Returns:
        _type_: function for splitting text based on the method provided 
    """
    if not isinstance(split_by, str):
        raise TypeError(f"expected str received {type(split_by)}")

    try:
        nltk.data.find("tokenizers/punkt")
    except LookupError:
        nltk.download("punkt")

    MAPPING = {
        'character':list,
        'word': partial(word_tokenize, language=language),
        'sentence':partial(sent_tokenize, language=language),
        'passage':lambda x:x.split("\n\n")
    }

    if split_by in MAPPING:
        return MAPPING[split_by]
    
    # can also have a custom split but leave that for now
    raise KeyError(f"unexpected split_by type of {split_by}")

def _reconstruct_single_list(segmented_text: List[str], seperator: str = " ") -> str:
    """_summary_

    Args:
        segmented_text (List[str]): List of strings that were segmented 
        seperator (str, optional): what to use as a seperator between tokens when re-joining into text. Defaults to " ".

    Returns:
        str: string of text that was re-joined
    """
    return seperator.join([t for t in segmented_text if t is not None])

def _reconstruct_multi_list(segmented_text_list: List[List[str]], seperator: str = " ") -> List[str]:
    """_summary_

    Args:
        segmented_text_list (List[List[str]]): _description_
        seperator (str, optional): _description_. Defaults to " ".

    Returns:
        List[str]: _description_
    """

    results = []
    for seg in segmented_text_list:
        txt = _reconstruct_single_list(seg, seperator)
        if len(txt) > 0:
            results.append(txt)
    
    return results

def check_make_string_valid(text: str, coerce: bool = True) -> str:
    """ does some simple validation and coercsion for empty strings

    Args:
        text (str): text of type str
        coerce (bool, optional): if an empty or None string is passed, return an empty string?. Defaults to True.

    Raises:
        TypeError: _description_

    Returns:
        _type_: something that is a string
    """
    empty_string = " "

    if text in [[], None, '', "", empty_string] and coerce:
        return empty_string

    if set(text) == set(" "):
        return empty_string

    if not isinstance(text, str):
        raise TypeError(f"text had type {type(text)} but expected str")

    return text

def split_text(text: str, split_by: str = 'sentence', split_length: int = 2, split_overlap: int = 1, 
               language: str = 'english', custom_seperator: str = None) -> List[str]:
    """ splits a single piece of text into smaller sub-texts based on splitting method (split_by).
        for example, the text can can be split at the character, word, sentence or passage level.
        optionally it can be split with a custom splitting string

    Args:
        text (str): _description_
        split_by (str, optional): _description_. Defaults to 'character'.
        split_length (int, optional): _description_. Defaults to 128.
        split_overlap (int, optional): _description_. Defaults to 16.
        language (str, optional): _description_. Defaults to 'english'.
        seperator (str, optional): _description_. Defaults to " ".

    Raises:
        TypeError: _description_

    Returns:
        List[str]: _description_
    """

    if split_length == 0:
        raise ValueError("split length must be > 0")

    # simple validation and correction    
    text = check_make_string_valid(text, coerce=True)

    # don't split if it is not worth splitting
    if len(text) <= 1:
        return [text]

    # we need to treat character splitting differently
    if custom_seperator is None:
        seperator = '' if split_by == 'character' else ' '
    else: 
        seperator = custom_seperator

    # determine how we want to split
    _func = _splitting_functions(split_by, language=language)

    # do the splitting
    split_text = _func(text)
    
    # concatenate individual elements based on split_length & split_stride
    segments = list(windowed(split_text, n=split_length, step=split_length - split_overlap))

    # reconstruct the segments. there is potential for a lossy process here as we
    # assume a uniform seperator when reconstructing the sentences
    text_splits = _reconstruct_multi_list(segments, seperator)

    return text_splits




"""neural search logic. In the future this will be accessible to the client via an API

API Notes:
    - Fields beginning with a double underscore "__" are protected and used for our internal purposes.
    - Examples include:
        __embedding_vector
        __field_name
        __field_content
        __doc_chunk_relation
        __chunk_ids
    - The "_id" field isn't a real field. It's a way to declare an ID. Internally we use it as the ID
        for the doc. The doc is stored without this field in its body

Notes on search behaviour with caching and searchable attributes:
    The behaviour of lexical search and vector search differs when it comes to
    interactions between the cache and searchable attributes.

    This issue should just occur on the first search when another user adds a
    new field, as the index cache updates in the background during the search.

    Lexical search:
        - Searching an existing but uncached field will return the best result
            (the uncached field will be searched)
        - Searching all fields will return a poor result
            (the uncached field won’t be searched)
    Vector search:
        - Searching an existing but uncached field will return no results (the
            uncached field won’t be searched)
        - Searching all fields will return a poor result (the uncached field
            won’t be searched)

"""
import copy
import datetime
import functools
import json
import pprint
import typing
import uuid
import asyncio
from typing import List, Optional, Union, Callable, Iterable, Sequence, Dict
import requests
from marqo.neural_search.enums import MediaType, MlModel, NeuralField, SearchMethod
from marqo.neural_search.enums import NeuralSettingsField as NsField
from marqo.neural_search import utils, backend, validation, configs
from marqo.processing import text as text_processor
from marqo.s2_inference import s2_inference
from marqo.neural_search.index_meta_cache import get_cache,get_index_info
from marqo.neural_search import index_meta_cache
from marqo.neural_search.models.index_info import IndexInfo
from marqo.neural_search import constants
# We depend on _httprequests.py for now, but this may be replaced in the future, as
# _httprequests.py is designed for the client
from marqo._httprequests import HttpRequests
from marqo.config import Config
from marqo.errors import MarqoError, MarqoApiError
import threading


def create_vector_index(
        config: Config, index_name: str, media_type: Union[str, MediaType] = MediaType.default,
        refresh_interval: str = "1s", number_of_shards=5, neural_settings = None):
    """
    Args:
        media_type: 'text'|'image'
    """

    vector_index_settings = {
        "settings": {
            "index": {
                "knn": True,
                "knn.algo_param.ef_search": 100,
                "refresh_interval":  refresh_interval
            },
            "number_of_shards": number_of_shards
        },
        "mappings": {
            "_meta": {
                "media_type": media_type,
            },
            "properties": {
                NeuralField.chunks: {
                    "type": "nested",
                    "properties": {
                        NeuralField.field_name: {
                            "type": "keyword"
                        },
                        NeuralField.field_content: {
                            "type": "text"
                        },
                    }
                }
            }
        }
    }

    if neural_settings is not None:
        the_neural_settings = _autofill_neural_settings(neural_settings=neural_settings)
    else:
        the_neural_settings = configs.get_default_neural_index_settings()

    model_name = the_neural_settings[NsField.index_defaults][NsField.model]
    vector_index_settings["mappings"]["_meta"][NsField.neural_settings] = the_neural_settings
    vector_index_settings["mappings"]["_meta"]["model"] = model_name

    response = HttpRequests(config).put(path=index_name, body=vector_index_settings)

    get_cache()[index_name] = IndexInfo(
        model_name=model_name, properties=vector_index_settings["mappings"]["properties"].copy(),
        neural_settings=the_neural_settings
    )
    return response


def _autofill_neural_settings(neural_settings: dict):
    """A half-complete neural settings will be auto filled"""

    # TODO: validated conflicting settings
    # treat_urls_and_pointers_as_images

    copied_settings = neural_settings.copy()
    default_settings = configs.get_default_neural_index_settings()

    if NsField.index_defaults not in copied_settings:
        copied_settings[NsField.index_defaults] = default_settings[NsField.index_defaults]

    if NsField.treat_urls_and_pointers_as_images in copied_settings[NsField.index_defaults] and \
            copied_settings[NsField.index_defaults][NsField.treat_urls_and_pointers_as_images] is True\
            and copied_settings[NsField.index_defaults][NsField.model] is None:
        copied_settings[NsField.index_defaults][NsField.model] = MlModel.clip

    for key in list(default_settings[NsField.index_defaults]):
        if key not in copied_settings[NsField.index_defaults] or \
                copied_settings[NsField.index_defaults][key] is None:
            copied_settings[NsField.index_defaults][key] = default_settings[NsField.index_defaults][key]

    for key in list(default_settings[NsField.index_defaults][NsField.text_preprocessing]):
        if key not in copied_settings[NsField.index_defaults][NsField.text_preprocessing] or \
                copied_settings[NsField.index_defaults][NsField.text_preprocessing][key] is None:
            copied_settings[NsField.index_defaults][NsField.text_preprocessing][key] \
                = default_settings[NsField.index_defaults][NsField.text_preprocessing][key]

    return copied_settings


def get_stats(config: Config, index_name: str):
    doc_count = HttpRequests(config).post(path=F"{index_name}/_count")["count"]
    return {
        "numberOfDocuments": doc_count
    }


def add_documents(config: Config, index_name: str, docs: List[dict], auto_refresh):
    """
    FIXME:
        - Atomicity issue: we index parent doc OK, but run into error indexing chunks.
            This means we have a parent doc unsearchable via vector search!
        - Designed for only a 1-shard index!
        - Appropriate error handler if chunks can't be added to index
    """

    bulk_parent_dicts = []

    try:
        index_info = backend.get_index_info(config=config, index_name=index_name)
    except MarqoApiError as s:
        if s.status_code == 404 and "index_not_found_exception" in str(s):
            create_vector_index(config=config, index_name=index_name)
            index_info = backend.get_index_info(config=config, index_name=index_name)
        else:
            raise s

    existing_fields = set(index_info.properties.keys())
    new_fields = set()
    doc_ids_to_update = []

    for doc in docs:

        indexing_instructions = {"index": {"_index": index_name}}
        copied = doc.copy()

        validation.validate_doc(doc)
        [validation.validate_field_name(field) for field in copied]

        if "_id" in doc:
            doc_id = doc["_id"]
            if not isinstance(doc_id, str):
                raise MarqoError("Document _id must be a string type! "
                                    f"Received _id {doc_id} of type {type(doc_id)}")
            del copied["_id"]
            doc_ids_to_update.append(doc_id)
        else:
            doc_id = str(uuid.uuid4())

        indexing_instructions["index"]["_id"] = doc_id

        chunks = []

        for field in copied:

            if field not in existing_fields:
                new_fields.add(field)

            field_content = copied[field]

            if isinstance(field_content, str):
                text_chunks = text_processor.split_text(field_content)
                vector_chunks = s2_inference.vectorise(index_info.model_name, text_chunks, 
                                                    config.indexing_device, index_info.neural_settings['index_defaults']['normalize_embeddings'],
                                                    infer=index_info.neural_settings[NsField.index_defaults][NsField.treat_urls_and_pointers_as_images])
                assert len(vector_chunks) == len(text_chunks)
                for text_chunk, vector_chunk in zip(text_chunks, vector_chunks):
                    chunk_id = str(uuid.uuid4())
                    chunks.append({
                        utils.generate_vector_name(field): vector_chunk,
                        NeuralField.field_content: text_chunk,
                        NeuralField.field_name: field,
                    })
        copied[NeuralField.chunks] = chunks
        bulk_parent_dicts.append(indexing_instructions)
        bulk_parent_dicts.append(copied)

    # the HttpRequest wrapper handles error logic
    update_mapping_response = backend.add_customer_field_properties(
        config=config, index_name=index_name, customer_field_names=new_fields,
        model_properties=s2_inference.get_model_properties(model_name=index_info.model_name))

    index_parent_response = HttpRequests(config).post(
        path="_bulk", body=utils.dicts_to_jsonl(bulk_parent_dicts))

    if auto_refresh:
        refresh_response = HttpRequests(config).post(path=F"{index_name}/_refresh")

    def translate_add_doc_response(response: dict) -> dict:
        """translates OpenSearch response dict into Marqo dict"""
        copied_res = copy.deepcopy(response)
        took = copied_res["took"]
        del copied_res["took"]
        new_items = []
        item_fields_to_remove = ['_index', '_primary_term', '_seq_no', '_shards', '_version']
        for item in copied_res["items"]:
            for to_remove in item_fields_to_remove:
                del item["index"][to_remove]
            new_items.append(item["index"])
        copied_res["processingTimeMs"] = took
        copied_res["index_name"] = index_name
        copied_res["items"] = new_items
        return copied_res

    return translate_add_doc_response(index_parent_response)


def get_document_by_id(config: Config, index_name:str, document_id: str):
    """returns document by its ID"""
    if not isinstance(document_id, str):
        raise MarqoError("Document ID must be a str")
    if not document_id:
        raise MarqoError("Document ID must can't be empty")
    res = HttpRequests(config).get(
        f'{index_name}/_doc/{document_id}'
    )
    if "_source" in res:
        return _clean_doc(res["_source"], doc_id=document_id)
    else:
        return res


def delete_documents(config: Config, index_name: str, doc_ids: List[str], auto_refresh):
    """Deletes documents """
    if not doc_ids:
        raise MarqoError("doc_ids can't be empty!")
    delete_parents_res = HttpRequests(config=config).post(
        path=f"{index_name}/_delete_by_query", body={
            "query": {
                "terms": {
                    "_id": doc_ids
                }
            }
        }
    )
    if auto_refresh:
        refresh_response = HttpRequests(config).post(path=F"{index_name}/_refresh")
    return delete_parents_res


def search(config: Config, index_name: str, text: str, result_count: int = 3, highlights=True, return_doc_ids=False,
           search_method: Union[str, SearchMethod, None] = SearchMethod.NEURAL,
           searchable_attributes: Iterable[str] = None, verbose=0, num_highlights=3) -> Dict:
    """The root search method. Calls the specific search method

    Validation should go here. Validations include:
        - all args and their types
        - result_count (negatives etc)
        - text

    This deals with index caching

    Args:
        config:
        index_name:
        text:
        result_count:
        return_doc_ids:
        search_method:
        searchable_attributes:
        verbose:
        num_highlights: number of highlights to return for each doc

    Returns:

    """
    MAX_RESULT_COUNT = 500

    if result_count > MAX_RESULT_COUNT or result_count < 0:
        raise MarqoError("result count must be between 0 and 500!")

    t0 = datetime.datetime.now()

    if searchable_attributes is not None:
        [validation.validate_field_name(attribute) for attribute in searchable_attributes]

    if verbose:
        print(f"determined_search_method: {search_method}, text query: {text}")
    # if we can't see the index name in cache, we request it and wait for the info
    if index_name not in index_meta_cache.get_cache():
        backend.get_index_info(config=config, index_name=index_name)

    # update cache in the background
    cache_update_thread = threading.Thread(
        target=index_meta_cache.refresh_index,
        args=(config, index_name))
    cache_update_thread.start()

    if search_method.upper() == SearchMethod.NEURAL:
        search_result = _vector_text_search(
            config=config, index_name=index_name, text=text, result_count=result_count,
            return_doc_ids=return_doc_ids, searchable_attributes=searchable_attributes,
            number_of_highlights=num_highlights
        )
    elif search_method.upper() == SearchMethod.LEXICAL:
        search_result = _lexical_search(
            config=config, index_name=index_name, text=text, result_count=result_count,
            return_doc_ids=return_doc_ids, searchable_attributes=searchable_attributes,
        )
    else:
        raise MarqoError(f"Search called with unknown search method: {search_method}")
    time_taken = datetime.datetime.now() - t0
    search_result["processingTimeMs"] = round(time_taken.total_seconds() * 1000)
    search_result["query"] = text
    search_result["limit"] = result_count

    if not highlights:
        for hit in search_result["hits"]:
            del hit["_highlights"]

    return search_result


def _lexical_search(
        config: Config, index_name: str, text: str, result_count: int = 3, return_doc_ids=False,
        searchable_attributes: Sequence[str] = None, raise_for_searchable_attributes=False):
    """

    Args:
        config:
        index_name:
        text:
        result_count:
        return_doc_ids:
        searchable_attributes:
        number_of_highlights:
        verbose:
        raise_for_searchable_attributes: if True, check searchable attributes and raises
            and error if identified. This is appropriate only if we assume that the cache
            remains in sync with the index. If that syncing assumption is lost, then this
            behaviour may be unexpected.

    Returns:

    Notes:
        Should not be directly called by client - the search() method should
        be called. The search() method adds syncing
    TODO:
        - Test raise_for_searchable_attribute=False
    """
    if not isinstance(text, str):
        raise MarqoError(f"text arg must be of type str! text arg is of type {type(text)}. "
                            f"text arg: {text}")

    if searchable_attributes is not None:
        cached_fields_to_search = [field for field in index_meta_cache.get_index_info(
                            config=config, index_name=index_name).get_text_properties()
                            if field in searchable_attributes]
        if raise_for_searchable_attributes and len(cached_fields_to_search) < len(searchable_attributes):
            raise MarqoError(
                "Detected unknown searchable_attributes:\n "
                f"Known attributes: {index_meta_cache.get_index_info(config=config, index_name=index_name).get_text_properties().keys()},\n"
                f"Searchable attributes: {searchable_attributes}")
        else:
            fields_to_search = searchable_attributes
    else:
        fields_to_search = index_meta_cache.get_index_info(
            config=config, index_name=index_name
        ).get_text_properties()

    search_res = HttpRequests(config).get(path=f"{index_name}/_search", body={
        "query": {
            "bool": {
                "should": [
                    {"match": {field: text}}
                    for field in fields_to_search
                ],
                "must_not": [
                    {"exists": {"field": NeuralField.field_name}}
                ]
            }
        },
        "size": result_count,
    })
    res_list = []
    for doc in search_res['hits']['hits']:
        just_doc = _clean_doc(doc["_source"].copy())
        if return_doc_ids:
            just_doc["_id"] = doc["_id"]
            just_doc["_score"] = doc["_score"]
        res_list.append({**just_doc, "_highlights": []})
    return {'hits': res_list[:result_count]}


def _vector_text_search(
        config: Config, index_name: str, text: str, result_count: int = 5, return_doc_ids=False,
        searchable_attributes: Iterable[str] = None, number_of_highlights=3,
        verbose=0, raise_on_searchable_attribs=False, hide_vectors=True, k=500,
        simplified_format=True
):
    """
    Args:
        config:
        index_name:
        text:
        result_count:
        return_doc_ids: if True adds doc _id to the docs. Otherwise just returns the docs as-is
        searchable_attributes: Iterable of field names to search. If left as None, then all will
            be searched
        number_of_highlights: if None, will return all highlights in
            descending order of relevancy. Otherwise will return this number of highlights
        verbose: if 0 - nothing is printed. if 1 - data is printed without vectors, if 2 - full
            objects are printed out
        hide_vectors: if True, vectors won't be returned from OpenSearch. This reduces the size
            of data transfers
    Returns:

    Note:
        - looks for k results in each attribute. Not that much of a concern unless you have a
        ridiculous number of attributes
        - Should not be directly called by client - the search() method should
        be called. The search() method adds syncing

    Output format:
        [
            {
                _id: doc_id
                doc: {# original document},
                highlights:[{}],
            },
        ]
    Future work:
        - max result count should be in a config somewhere
        - searching a non existent index should return a HTTP-type error
    """
    if result_count < 0 or result_count > constants.MAX_VECTOR_SEARCH_RESULT_COUNT:
        raise ValueError("neural_search: vector_text_search: illegal result_count: {}".format(result_count))

    try:
        index_info = get_index_info(config=config, index_name=index_name)
    except KeyError as e:
        raise MarqoError(message="Tried to search a non-existent index: {}".format(index_name))
    vectorised_text = s2_inference.vectorise(
        model_name=index_info.model_name, content=text_processor.split_text(text)[0], 
        device=config.search_device,
        normalize_embeddings=index_info.neural_settings['index_defaults']['normalize_embeddings'])[0]

    body = []

    if searchable_attributes is None:
        vector_properties_to_search = index_info.get_vector_properties().keys()
    else:
        if raise_on_searchable_attribs:
            vector_properties_to_search = validation.validate_searchable_vector_props(
                existing_vector_properties=index_info.get_vector_properties().keys(),
                subset_vector_properties=searchable_attributes
            )
        else:
            searchable_attributes_as_vectors = {utils.generate_vector_name(field_name=attribute)
                                                for attribute in searchable_attributes}
            # discard searchable attributes that aren't found in the cache:
            vector_properties_to_search = searchable_attributes_as_vectors.intersection(
                index_info.get_vector_properties().keys())

    for vector_field in vector_properties_to_search:
        search_query = {
            "size": result_count,
            "query": {
                "nested": {
                    "path": NeuralField.chunks,
                    "inner_hits": {
                        "_source": {
                            "exclude": ["*__vector*"]
                        }
                    },
                    "query": {
                        "knn": {
                            f"{NeuralField.chunks}.{vector_field}": {
                                "vector": vectorised_text,
                                "k": k
                            }
                        }
                    },
                    "score_mode": "max",
                }
            }
        }
        if hide_vectors:
            search_query["_source"] = {
                "exclude": ["*__vector*"]
            }
            search_query["query"]["nested"]["inner_hits"]["_source"] = {
                "exclude": ["*__vector*"]
            }
        body += [{"index": index_name}, search_query]

    if verbose:
        print("vector search body:")
        if verbose == 1:
            readable_body = copy.deepcopy(body)
            for i, q in enumerate(readable_body):
                if "index" in q:
                    continue
                for vec in list(q["query"]["nested"]["query"]["knn"].keys()):
                    readable_body[i]["query"]["nested"]["query"]["knn"][vec]["vector"] = readable_body[i]["query"]["nested"]["query"]["knn"][vec]["vector"][:5]
            pprint.pprint(readable_body)
        if verbose == 2:
            pprint.pprint(body, compact=True)

    if not body:
        # empty body means that there are no vector fields associated with the index.
        # This probably means the index is emtpy
        return {"hits": []}
    response = HttpRequests(config).get(path=F"{index_name}/_msearch", body=utils.dicts_to_jsonl(body))

    responses = [r['hits']['hits'] for r in response["responses"]]
    gathered_docs = dict()

    if verbose:
        print("search responses:")
        pprint.pprint(responses)
    for i, query_res in enumerate(responses):
        for doc in query_res:
            doc_chunks = doc["inner_hits"][NeuralField.chunks]["hits"]["hits"]
            if doc["_id"] in gathered_docs:
                gathered_docs[doc["_id"]]["doc"] = doc
                gathered_docs[doc["_id"]]["chunks"].extend(doc_chunks)
            else:
                gathered_docs[doc["_id"]] = {
                    "_id": doc["_id"],
                    "doc": doc,
                    "chunks": doc_chunks
                }

    # Filter out docs with no inner hits:

    for doc_id in list(gathered_docs.keys()):
        if not gathered_docs[doc_id]["chunks"]:
            del gathered_docs[doc_id]

    # SORT THE DOCS HERE

    def sort_chunks(docs: dict) -> dict:
        to_be_sorted = docs.copy()
        for doc_id in list(to_be_sorted.keys()):
            to_be_sorted[doc_id]["chunks"] = sorted(
                to_be_sorted[doc_id]["chunks"], key=lambda x: x["_score"], reverse=True)
        return to_be_sorted

    docs_chunks_sorted = sort_chunks(gathered_docs)

    def sort_docs(docs: dict) -> List[dict]:
        as_list = list(docs.values())
        return sorted(as_list,  key=lambda x: x["chunks"][0]["_score"], reverse=True)

    completely_sorted = sort_docs(docs_chunks_sorted)

    if verbose:
        print("Chunk vector search, sorted result:")
        if verbose == 1:
            pprint.pprint(utils.truncate_dict_vectors(completely_sorted))
        elif verbose == 2:
            pprint.pprint(completely_sorted)

    # format output:
    def format_ordered_docs_preserving(ordered_docs_w_chunks: List[dict], num_highlights: Optional[int]) -> dict:
        """Formats docs so that it preserves the original document, unless doc_ids are returned
        Args:
            ordered_docs_w_chunks:
            num_highlights: number of highlights to return.
        Returns:
        """
        return {'hits': [dict([
            ('doc', _clean_doc(doc['doc']["_source"], doc_id=doc['_id'] if return_doc_ids else None)),
            ('highlights', [{
                    the_chunk["_source"][NeuralField.field_name]: the_chunk["_source"][NeuralField.field_content]
                } for the_chunk in doc['chunks']][:num_highlights])
        ]) for doc in ordered_docs_w_chunks][:result_count]}

    # format output:
    def format_ordered_docs_simple(ordered_docs_w_chunks: List[dict]) -> dict:
        """Only one highlight is returned
        Args:
            ordered_docs_w_chunks:
        Returns:
        """
        simple_results = []

        for d in ordered_docs_w_chunks:
            cleaned = _clean_doc(d['doc']["_source"], doc_id=d['_id'])
            cleaned["_highlights"] = {
                d["chunks"][0]["_source"][NeuralField.field_name]: d["chunks"][0]["_source"][NeuralField.field_content]
            }
            cleaned["_score"] = d["chunks"][0]["_score"]
            simple_results.append(cleaned)
        return {"hits": simple_results[:result_count]}

    if simplified_format:
        return format_ordered_docs_simple(ordered_docs_w_chunks=completely_sorted)
    else:
        return format_ordered_docs_preserving(ordered_docs_w_chunks=completely_sorted, num_highlights=number_of_highlights)


def delete_index(config: Config, index_name):
    res = HttpRequests(config).delete(path=index_name)
    if index_name in get_cache():
        del get_cache()[index_name]
    return res


def _select_vectoriser(model_name: Union[str, MlModel]) -> Callable[[str], List[List[float]]]:
    """selects the vectorisation function based on the model's name"""
    if model_name == MlModel.bert or model_name == MlModel.clip:
        return functools.partial(s2_inference.vectorise, model_name)
    else:
        raise ValueError("neural_search: Unknown model selected: {}".format(model_name))


def _clean_doc(doc: dict, doc_id=None) -> dict:
    """clears neural search specific fields from the doc

    Args:
        doc: the doc to clean
        doc_id: if left as None, then the doc will be returned without the _id field

    Returns:

    """
    copied = doc.copy()
    if NeuralField.doc_chunk_relation in copied:
        del copied[NeuralField.doc_chunk_relation]
    if NeuralField.chunk_ids in copied:
        del copied[NeuralField.chunk_ids]
    if NeuralField.chunks in copied:
        del copied[NeuralField.chunks]
    if doc_id is not None:
        copied['_id'] = doc_id
    return copied


def _select_model_from_media_type(media_type: Union[MediaType, str]) -> Union[MlModel, str]:
    if media_type == MediaType.text:
        return MlModel.bert
    elif media_type == MediaType.image:
        return MlModel.clip
    else:
        raise ValueError("_select_model_from_media_type(): "
                         "Received unknown media type: {}".format(media_type))






import torch
from torch import nn
from transformers import (AutoModel, AutoTokenizer)
import numpy as np

from marqo.s2_inference.sbert_utils import Model
from marqo.s2_inference.types import Union, FloatTensor, List

from marqo.s2_inference.logger import get_logger
logger = get_logger(__name__)


class HF_MODEL(Model):

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        if self.max_seq_length is None:
            self.max_seq_length = 128

    def load(self) -> None:

        self.model = AutoModelForSentenceEmbedding(self.model_name).to(self.device)
        self.tokenizer = AutoTokenizer.from_pretrained(self.model_name)


    def encode(self, sentence: Union[str, List[str]], normalize = True, **kwargs) -> Union[FloatTensor, np.ndarray]:
        
        if isinstance(sentence, str):
            sentence = [sentence]

        if self.model is None:
            self.load()

        self.model.normalize = normalize

        inputs = self.tokenizer(sentence, padding=True, truncation=True, max_length=self.max_seq_length, return_tensors="pt").to(self.device)

        with torch.no_grad():
            return self.model.forward(**inputs)

class AutoModelForSentenceEmbedding(nn.Module):

    def __init__(self, model_name, normalize = True, pooling='mean'):
        super().__init__()
        self.model_name = model_name
        self.normalize = normalize
        self.pooling = pooling

        self.model = AutoModel.from_pretrained(model_name)
        self.model.eval()
        if self.pooling == 'mean':
            self._pool_func = self.mean_pooling
        elif self.pooling == 'cls':
            self._pool_func = self.cls_pooling
        else:
            raise TypeError(f"{pooling} not in allowed pooling types of 'mean' or 'cls' ")

    def forward(self, **kwargs):

        model_output = self.model(**kwargs)

        embeddings = self._pool_func(model_output, kwargs['attention_mask'])

        if self.normalize:
            return nn.functional.normalize(embeddings, p=2, dim=1)

        return embeddings

    def mean_pooling(self, model_output, attention_mask):

        token_embeddings = model_output[0]
        
        input_mask_expanded = attention_mask.unsqueeze(-1).expand(token_embeddings.size()).float()

        return torch.sum(token_embeddings * input_mask_expanded, 1) / torch.clamp(input_mask_expanded.sum(1), min=1e-9)

    def cls_pooling(self, model_output, attention_mask):
        return model_output[0][:,0]